"""Functions to raise warnings based on user inputs."""

from warnings import warn

from .connections import ConnectionManager, TcpConnectionManager
from .requests import Request


def raise_warnings(request: Request, connection: ConnectionManager):
    """Calls all warning functions.

    :param request: Instance of a request.
    :param connection: Connection manager instance.
    """

    warn_spamd_bug_7183(request, connection)


def warn_spamd_bug_7183(request: Request, connection: ConnectionManager):
    """Warn on spamd bug if using compression with an SSL connection.

    :param request: Instance of a request.
    :param connection: Connection manager instance.

    Bug: https://bz.apache.org/SpamAssassin/show_bug.cgi?id=7183
    """

    if (
        "Compress" in request.headers
        and isinstance(connection, TcpConnectionManager)
        and connection.ssl_context is not None
    ):
        message = (
            "spamd bug 1783: SpamAssassin hangs when using SSL and compression are used in combination. "
            "Disable compression as a workaround. "
            "More information available at: https://bz.apache.org/SpamAssassin/show_bug.cgi?id=7183"
        )
        warn(message)



"""Collection of request and response header value objects."""

import getpass
from base64 import b64encode
from collections import UserDict
from dataclasses import dataclass
from enum import Enum
from typing import Any, Optional, Protocol


class HeaderValue(Protocol):  # pragma: no cover
    """Protocol for headers."""

    def __bytes__(self) -> bytes:
        pass

    def to_json(self) -> Any:
        """Convert to a JSON object."""

        pass


@dataclass
class BytesHeaderValue:
    """Header with bytes value.

    :param value: Value of the header.
    """

    value: bytes

    def __bytes__(self) -> bytes:
        return self.value

    def to_json(self) -> Any:
        """Converts object to a JSON serializable object."""

        return b64encode(self.value).decode()


@dataclass
class GenericHeaderValue:
    """Generic header value."""

    value: str
    encoding: str = "utf8"

    def __bytes__(self) -> bytes:
        return self.value.encode(self.encoding)

    def to_json(self) -> Any:
        """Converts object to a JSON serializable object."""

        return self.value


@dataclass
class CompressValue:
    """Compress header.  Specifies what encryption scheme to use.  So far only
    'zlib' is supported.
    """

    algorithm: str = "zlib"

    def __bytes__(self) -> bytes:
        return self.algorithm.encode("ascii")

    def to_json(self) -> Any:
        """Converts object to a JSON serializable object."""

        return self.algorithm


@dataclass
class ContentLengthValue:
    """ContentLength header.  Indicates the length of the body in bytes."""

    length: int = 0

    def __int__(self) -> int:
        return self.length

    def __bytes__(self) -> bytes:
        return str(self.length).encode("ascii")

    def to_json(self) -> Any:
        """Converts object to a JSON serializable object."""

        return self.length


class MessageClassOption(str, Enum):
    """Option to be used for the MessageClass header."""

    spam = "spam"
    ham = "ham"

    def __str__(self) -> str:
        return self.value


@dataclass
class MessageClassValue:
    """MessageClass header.  Used to specify whether a message is 'spam' or
    'ham.'
    """

    value: MessageClassOption = MessageClassOption.ham

    def __bytes__(self) -> bytes:
        return self.value.name.encode("ascii")

    def to_json(self) -> Any:
        """Converts object to a JSON serializable object."""

        return self.value.value


@dataclass
class ActionOption:
    """Option to be used in the DidRemove, DidSet, Set, and Remove headers.

    :param local: An action will be performed on the SPAMD service's local database.
    :param remote: An action will be performed on the SPAMD service's remote database.
    """

    local: bool = False
    remote: bool = False


@dataclass
class SetOrRemoveValue:
    """Base class for headers that implement "local" and "remote" rules."""

    action: ActionOption

    def __bytes__(self) -> bytes:
        if not self.action.local and not self.action.remote:
            # if nothing is set, then return a blank string so the request
            # doesn't get tainted
            return b""

        values = []
        if self.action.local:
            values.append(b"local")
        if self.action.remote:
            values.append(b"remote")

        return b", ".join(values)

    def to_json(self) -> Any:
        """Converts object to a JSON serializable object."""

        return {"local": self.action.local, "remote": self.action.remote}


@dataclass
class SpamValue:
    """Spam header.  Used by the SPAMD service to report on if the submitted
    message was spam and the score/threshold that it used."""

    value: bool = False
    score: float = 0.0
    threshold: float = 0.0

    def __bytes__(self) -> bytes:
        return b"%b ; %.1f / %.1f" % (
            str(self.value).encode("ascii"),
            self.score,
            self.threshold,
        )

    def to_json(self) -> Any:
        """Converts object to a JSON serializable object."""

        return {"value": self.value, "score": self.score, "threshold": self.threshold}


@dataclass
class UserValue:
    """User header.  Used to specify which user the SPAMD service should use
    when loading configuration files."""

    name: str = getpass.getuser()

    def __bytes__(self) -> bytes:
        return self.name.encode("ascii")

    def __str__(self) -> str:
        return self.name

    def to_json(self) -> Any:
        """Converts object to a JSON serializable object."""

        return self.name


class Headers(UserDict):
    """Class to store headers with shortcut properties."""

    def get_header(self, name: str) -> Optional[str]:
        """Get a string header if it exists.

        :param name: Name of the header.
        :return: The header value.
        """

        if header := self.data.get(name):
            return header.value
        return None

    def set_header(self, name: str, value: str):
        """Sets a string header.

        :param name: Name of the header.
        :param value: Value of the header.
        """

        self.data[name] = GenericHeaderValue(value)

    def get_bytes_header(self, name: str) -> Optional[bytes]:
        """Get a bytes header if it exists.

        :param name: Name of the header.
        :return: The header value.
        """

        if header := self.data.get(name):
            return header.value
        return None

    def set_bytes_header(self, name: str, value: bytes):
        """Sets a string header.

        :param name: Name of the header.
        :param value: Value of the header.
        """

        self.data[name] = BytesHeaderValue(value)

    @property
    def compress(self) -> Optional[str]:
        """Gets the Compress header if it exists.

        :return: Compress header value.
        """

        if header := self.data.get("Compress"):
            return header.algorithm
        return None

    @compress.setter
    def compress(self, value: str = "zlib"):
        """Sets the Compress header.

        :param value: Value of the header.
        """

        self.data["Compress"] = CompressValue(value)

    @property
    def content_length(self) -> Optional[int]:
        """Gets the Content-length header if it exists.

        :return: Content-length header value.
        """

        if header := self.data.get("Content-length"):
            return header.length
        return None

    @content_length.setter
    def content_length(self, value: int):
        """Sets the Content-length header.

        :param value: Value of the header.
        """

        self.data["Content-length"] = ContentLengthValue(value)

    @property
    def message_class(self) -> Optional[MessageClassOption]:
        """Gets the Message-class header if it exists.

        :return: Message-class header value.
        """

        if header := self.data.get("Message-class"):
            return header.value
        return None

    @message_class.setter
    def message_class(self, value: MessageClassOption):
        """Sets the Message-class header.

        :param value: Value of the header.
        """

        self.data["Message-class"] = MessageClassValue(value)

    @property
    def set_(self) -> Optional[ActionOption]:
        """Gets the Set header if it exists.

        :return: Set header value.
        """

        if header := self.data.get("Set"):
            return header.action
        return None

    @set_.setter
    def set_(self, value: ActionOption):
        """Sets the Set header.

        :param value: Value of the header.
        """

        self.data["Set"] = SetOrRemoveValue(value)

    @property
    def remove(self) -> Optional[ActionOption]:
        """Gets the Remove header if it exists.

        :return: Remove header value.
        """

        if header := self.data.get("Remove"):
            return header.action
        return None

    @remove.setter
    def remove(self, value: ActionOption):
        """Sets the Remove header.

        :param value: Value of the header.
        """

        self.data["Remove"] = SetOrRemoveValue(value)

    @property
    def did_set(self) -> Optional[ActionOption]:
        """Gets the DidSet header if it exists.

        :return: DidSet header value.
        """

        if header := self.data.get("DidSet"):
            return header.action
        return None

    @did_set.setter
    def did_set(self, value: ActionOption):
        """Sets the DidSet header.

        :param value: Value of the header.
        """

        self.data["DidSet"] = SetOrRemoveValue(value)

    @property
    def did_remove(self) -> Optional[ActionOption]:
        """Gets the DidRemove header if it exists.

        :return: DidRemove header value.
        """

        if header := self.data.get("DidRemove"):
            return header.action
        return None

    @did_remove.setter
    def did_remove(self, value: ActionOption):
        """Sets the DidRemove header.

        :param value: Value of the header.
        """

        self.data["DidRemove"] = SetOrRemoveValue(value)

    @property
    def spam(self) -> Optional[SpamValue]:
        """Gets the Spam header if it exists.

        :return: Spam header value.
        """

        return self.data.get("Spam")

    @spam.setter
    def spam(self, value: SpamValue):
        """Sets the Spam header.

        :param value: Value of the header.
        """

        self.data["Spam"] = value

    @property
    def user(self) -> Optional[str]:
        """Gets the User header if it exists.

        :return: User header value.
        """

        if header := self.data.get("User"):
            return header.name
        return None

    @user.setter
    def user(self, value: str):
        """Sets the User header.

        :param value: Value of the header.
        """

        self.data["User"] = UserValue(value)



"""Module implementing client objects that all requests go through."""

from loguru import logger

from .connections import ConnectionManager
from .exceptions import BadResponse
from .incremental_parser import ParseError, ResponseParser
from .requests import Request
from .responses import Response
from .user_warnings import raise_warnings


class Client:
    """Client object to submit requests."""

    def __init__(self, connection_manager: ConnectionManager):
        """Client constructor.

        :param connection_manager: Instance of a connection manager.
        """

        self.connection_manager = connection_manager

    async def request(self, req: Request):
        """Sends a request and returns the parsed response.

        :param req: The request to send.

        :return: The parsed response.

        :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
        :raises AIOSpamcConnectionFailed: Raised if an error occurred when trying to connect.
        :raises UsageException: Error in command line usage.
        :raises DataErrorException: Error with data format.
        :raises NoInputException: Cannot open input.
        :raises NoUserException: Addressee unknown.
        :raises NoHostException: Hostname unknown.
        :raises UnavailableException: Service unavailable.
        :raises InternalSoftwareException: Internal software error.
        :raises OSErrorException: System error.
        :raises OSFileException: Operating system file missing.
        :raises CantCreateException: Cannot create output file.
        :raises IOErrorException: Input/output error.
        :raises TemporaryFailureException: Temporary failure, may reattempt.
        :raises ProtocolException: Error in the protocol.
        :raises NoPermissionException: Permission denied.
        :raises ConfigException: Error in configuration.
        :raises ServerTimeoutException: Server returned a response that it timed out.
        :raises ClientTimeoutException: Client timed out during connection.
        """

        context_logger = logger.bind(
            connection=self.connection_manager,
            request=req,
        )

        raise_warnings(req, self.connection_manager)

        context_logger.info("Sending {} request", req.verb)
        response = await self.connection_manager.request(bytes(req))
        context_logger = context_logger.bind(response_bytes=response)
        try:
            parser = ResponseParser()
            parsed_response = parser.parse(response)
        except ParseError as error:
            context_logger.exception("Error parsing response")
            raise BadResponse(response) from error
        response_obj = Response(**parsed_response)
        response_obj.raise_for_status()
        context_logger.bind(response=response_obj).success(
            "Successfully received response"
        )

        return response_obj



"""ConnectionManager classes for TCP and Unix sockets."""

from __future__ import annotations

import asyncio
import ssl
from enum import Enum, auto
from getpass import getpass
from pathlib import Path
from typing import Any, Callable, Optional, Union

import certifi
import loguru
from loguru import logger

from .exceptions import AIOSpamcConnectionFailed, ClientTimeoutException


class Timeout:
    """Container object for defining timeouts."""

    def __init__(
        self,
        total: float = 600,
        connection: Optional[float] = None,
        response: Optional[float] = None,
    ) -> None:
        """Timeout constructor.

        :param total: The total length of time in seconds to set the timeout.
        :param connection: The length of time in seconds to allow for a connection to live before timing out.
        :param response: The length of time in seconds to allow for a response from the server before timing out.
        """

        self.total = float(total)
        self.connection = connection
        self.response = response

    def __repr__(self):
        return (
            f"{self.__class__.__qualname__}("
            f"total={self.total}, "
            f"connection={self.connection}, "
            f"response={self.response}"
            ")"
        )


class ConnectionManagerBuilder:
    """Builder for connection managers."""

    class ManagerType(Enum):
        """Define connection manager type during build."""

        Undefined = auto()
        Tcp = auto()
        Unix = auto()

    def __init__(self):
        """ConnectionManagerBuilder constructor."""

        self._manager_type = self.ManagerType.Undefined
        self._tcp_builder = TcpConnectionManagerBuilder()
        self._unix_builder = UnixConnectionManagerBuilder()
        self._ssl_builder = SSLContextBuilder()
        self._ssl = False
        self._timeout = None

    def build(self) -> Union[UnixConnectionManager, TcpConnectionManager]:
        """Builds the :class:`aiospamc.connections.ConnectionManager`.

        :return: An instance of :class:`aiospamc.connections.TcpConnectionManager`
            or :class:`aiospamc.connections.UnixConnectionManager`
        """

        if self._manager_type is self.ManagerType.Undefined:
            raise ValueError(
                "Connection type is undefined, builder must be called with 'with_unix_socket' or 'with_tcp'"
            )
        elif self._manager_type is self.ManagerType.Tcp:
            ssl_context = None if not self._ssl else self._ssl_builder.build()
            self._tcp_builder.set_ssl_context(ssl_context)
            return self._tcp_builder.set_timeout(self._timeout).build()
        else:
            return self._unix_builder.set_timeout(self._timeout).build()

    def with_unix_socket(self, path: Path) -> ConnectionManagerBuilder:
        """Configures the builder to use a Unix socket connection.

        :param path: Path to the Unix socket.

        :return: This builder instance.
        """

        self._manager_type = self.ManagerType.Unix
        self._unix_builder.set_path(path)
        self._tcp_host = self._tcp_port = None

        return self

    def with_tcp(self, host: str, port: int = 783) -> ConnectionManagerBuilder:
        """Configures the builder to use a TCP connection.

        :param host: Hostname to use.
        :param port: Port to use.

        :return: This builder instance.
        """

        self._manager_type = self.ManagerType.Tcp
        self._tcp_builder.set_host(host).set_port(port)
        self._unix_path = None

        return self

    def add_ssl_context(self, context: ssl.SSLContext) -> ConnectionManagerBuilder:
        """Adds an SSL context when a TCP connection is being used.

        :param context: :class:`ssl.SSLContext` instance.

        :return: This builder instance.
        """

        self._ssl_builder.with_context(context)
        self._ssl = True

        return self

    def set_timeout(self, timeout: Timeout) -> ConnectionManagerBuilder:
        """Sets the timeout for the connection.

        :param timeout: Timeout object.

        :return: This builder instance.
        """

        self._timeout = timeout

        return self


class ConnectionManager:
    """Stores connection parameters and creates connections."""

    def __init__(
        self, connection_string: str, timeout: Optional[Timeout] = None
    ) -> None:
        """ConnectionManager constructor.

        :param timeout: Timeout configuration
        """

        self._connection_string = connection_string
        self.timeout = timeout or Timeout()
        self._logger = logger.bind(
            connection_string=self.connection_string,
            timeout=self.timeout,
        )

    @property
    def logger(self) -> loguru.Logger:
        """Return the logger object."""

        return self._logger

    async def request(self, data: bytes) -> bytes:
        """Send bytes data and receive a response.

        :raises: AIOSpamcConnectionFailed
        :raises: ClientTimeoutException

        :param data: Data to send.
        """

        try:
            response = await asyncio.wait_for(self._send(data), self.timeout.total)
        except asyncio.TimeoutError:
            self.logger.exception("Total timeout reached")
            raise

        return response

    async def _send(self, data: bytes) -> bytes:
        """Opens a connection, sends data to the writer, waits for the reader, then returns the response.

        :param data: Data to send.

        :return: Byte data from the response.
        """

        reader, writer = await self._connect()

        writer.write(data)
        if writer.can_write_eof():
            writer.write_eof()
        await writer.drain()

        response = await self._receive(reader)

        writer.close()
        await writer.wait_closed()

        return response

    async def _receive(self, reader: asyncio.StreamReader) -> bytes:
        """Takes a reader and returns the response.

        :param reader: asyncio reader.

        :return: Byte data from the response.
        """

        try:
            response = await asyncio.wait_for(reader.read(), self.timeout.response)
        except asyncio.TimeoutError as error:
            self.logger.exception("Timed out receiving data")
            raise ClientTimeoutException from error

        self.logger.success("Successfully received data")

        return response

    async def _connect(self) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        """Opens a connection from the connection manager.

        :return: Tuple or asyncio reader and writer.
        """

        try:
            reader, writer = await asyncio.wait_for(
                self.open(), self.timeout.connection
            )
        except asyncio.TimeoutError as error:
            self.logger.exception("Timeout when connecting")
            raise ClientTimeoutException from error

        self.logger.success("Successfully connected")

        return reader, writer

    async def open(self) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        """Opens a connection, returning the reader and writer objects."""

        raise NotImplementedError

    @property
    def connection_string(self) -> str:
        """String representation of the connection."""

        return self._connection_string


class TcpConnectionManagerBuilder:
    """Builder for :class:`aiospamc.connections.TcpConnectionManager`"""

    def __init__(self):
        """`TcpConnectionManagerBuilder` constructor."""

        self._args = {}

    def build(self) -> TcpConnectionManager:
        """Builds the :class:`aiospamc.connections.TcpConnectionManager`.

        :return: An instance of :class:`aiospamc.connections.TcpConnectionManager`.
        """

        return TcpConnectionManager(**self._args)

    def set_host(self, host: str) -> TcpConnectionManagerBuilder:
        """Sets the host to use.

        :param host: Hostname to use.

        :return: This builder instance.
        """

        self._args["host"] = host
        return self

    def set_port(self, port: int) -> TcpConnectionManagerBuilder:
        """Sets the port to use.

        :param port: Port to use.

        :return: This builder instance.
        """

        self._args["port"] = port
        return self

    def set_ssl_context(self, context: ssl.SSLContext) -> TcpConnectionManagerBuilder:
        """Set an SSL context.

        :param context: An instance of :class:`ssl.SSLContext`.

        :return: This builder instance.
        """

        self._args["ssl_context"] = context
        return self

    def set_timeout(self, timeout: Timeout) -> TcpConnectionManagerBuilder:
        """Sets the timeout for the connection.

        :param timeout: Timeout object.

        :return: This builder instance.
        """

        self._args["timeout"] = timeout
        return self


class TcpConnectionManager(ConnectionManager):
    """Connection manager for TCP connections."""

    def __init__(
        self,
        host: str,
        port: int,
        ssl_context: Optional[ssl.SSLContext] = None,
        timeout: Optional[Timeout] = None,
    ) -> None:
        """TcpConnectionManager constructor.

        :param host: Hostname or IP address.
        :param port: TCP port.
        :param ssl_context: SSL context.
        :param timeout: Timeout configuration.
        """

        super().__init__(f"{host}:{port}", timeout)
        self.host = host
        self.port = port
        self.ssl_context = ssl_context

    async def open(self) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        """Opens a TCP connection.

        :raises: AIOSpamcConnectionFailed

        :return: Reader and writer for the connection.
        """

        try:
            reader, writer = await asyncio.open_connection(
                self.host, self.port, ssl=self.ssl_context
            )
        except (ConnectionRefusedError, OSError) as error:
            self.logger.exception("Exception occurred when connecting")
            raise AIOSpamcConnectionFailed from error

        return reader, writer


class UnixConnectionManagerBuilder:
    """Builder for :class:`aiospamc.connections.UnixConnectionManager`."""

    def __init__(self):
        """`UnixConnectionManagerBuilder` constructor."""

        self._args = {}

    def build(self) -> UnixConnectionManager:
        """Builds a :class:`aiospamc.connections.UnixConnectionManager`.

        :return: An instance of :class:`aiospamc.connections.UnixConnectionManager`.
        """

        return UnixConnectionManager(**self._args)

    def set_path(self, path: Path) -> UnixConnectionManagerBuilder:
        """Sets the unix socket path.

        :param path: Path to the Unix socket.

        :return: This builder instance.
        """

        self._args["path"] = path
        return self

    def set_timeout(self, timeout: Timeout) -> UnixConnectionManagerBuilder:
        """Sets the timeout for the connection.

        :param timeout: Timeout object.

        :return: This builder instance.
        """

        self._args["timeout"] = timeout
        return self


class UnixConnectionManager(ConnectionManager):
    """Connection manager for Unix pipes."""

    def __init__(self, path: Path, timeout: Optional[Timeout] = None):
        """UnixConnectionManager constructor.

        :param path: Unix socket path.
        :param timeout: Timeout configuration
        """

        super().__init__(str(path), timeout)
        self.path = path

    async def open(self) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        """Opens a unix socket path connection.

        :raises: AIOSpamcConnectionFailed

        :return: Reader and writer for the connection.
        """

        try:
            reader, writer = await asyncio.open_unix_connection(self.path)
        except (ConnectionRefusedError, OSError) as error:
            self.logger.exception("Exception occurred when connecting")
            raise AIOSpamcConnectionFailed from error

        return reader, writer


class SSLContextBuilder:
    """SSL context builder."""

    def __init__(self):
        """Builder contstructor. Sets up a default SSL context."""

        self._context = ssl.create_default_context()

    def build(self) -> ssl.SSLContext:
        """Builds the SSL context.

        :return: An instance of :class:`ssl.SSLContext`.
        """

        return self._context

    def with_context(self, context: ssl.SSLContext) -> SSLContextBuilder:
        """Use the SSL context.

        :param context: Provided SSL context.

        :return: The builder instance.
        """

        self._context = context

        return self

    def add_ca_file(self, file: Path) -> SSLContextBuilder:
        """Add certificate authority from a file.

        :param file: File of concatenated certificates.

        :return: The builder instance.
        """

        self._context.load_verify_locations(cafile=file)

        return self

    def add_ca_dir(self, dir: Path) -> SSLContextBuilder:
        """Add certificate authority from a directory.

        :param dir: Directory of certificates.

        :return: The builder instance.
        """

        self._context.load_verify_locations(capath=dir)

        return self

    def add_ca(self, path: Path) -> SSLContextBuilder:
        """Add a certificate authority.

        :param path: Directory or file of certificates.

        :return: The builder instance.
        """

        if path.is_dir():
            return self.add_ca_dir(path)
        elif path.is_file():
            return self.add_ca_file(path)
        else:
            raise FileNotFoundError(path)

    def add_default_ca(self) -> SSLContextBuilder:
        """Add default certificate authorities.

        :return: The builder instance.
        """

        self._context.load_verify_locations(cafile=certifi.where())

        return self

    def add_client(
        self,
        file: Path,
        key: Optional[Path] = None,
        password: Optional[Callable[[], Union[str, bytes, bytearray]]] = None,
    ) -> SSLContextBuilder:
        """Add client certificate.

        :param file: Path to the client certificate.
        :param key: Path to the key.
        :param password: Callable that returns the password, if any.
        """

        self._context.load_cert_chain(file, key, password)

        return self

    def dont_verify(self) -> SSLContextBuilder:
        """Set the context to not verify certificates."""

        self._context.check_hostname = False
        self._context.verify_mode = ssl.CERT_NONE

        return self



"""Contains classes used for responses."""

from __future__ import annotations

import zlib
from base64 import b64encode
from enum import IntEnum
from typing import Any, SupportsBytes, Union

from .exceptions import TimeoutException
from .header_values import ContentLengthValue, Headers


class Status(IntEnum):
    """Enumeration for the status values defined by SPAMD."""

    EX_OK = 0
    EX_USAGE = 64
    EX_DATAERR = 65
    EX_NOINPUT = 66
    EX_NOUSER = 67
    EX_NOHOST = 68
    EX_UNAVAILABLE = 69
    EX_SOFTWARE = 70
    EX_OSERR = 71
    EX_OSFILE = 72
    EX_CANTCREAT = 73
    EX_IOERR = 74
    EX_TEMPFAIL = 75
    EX_PROTOCOL = 76
    EX_NOPERM = 77
    EX_CONFIG = 78
    EX_TIMEOUT = 79


class Response:
    """Class to encapsulate response."""

    def __init__(
        self,
        version: str = "1.5",
        status_code: Union[Status, int] = 0,
        message: str = "",
        headers: Union[dict[str, Any], Headers, None] = None,
        body: bytes = b"",
        **_,
    ):
        """Response constructor.

        :param version: Version reported by the SPAMD service response.
        :param status_code: Success or error code.
        :param message: Message associated with status code.
        :param body: Byte string representation of the body.
        :param headers: Collection of headers to be added.
        """

        self.version = version
        if isinstance(headers, dict):
            self.headers = Headers(headers)
        elif isinstance(headers, Headers):
            self.headers = headers
        else:
            self.headers = Headers()
        self._status_code: Union[Status, int]
        self.status_code = status_code
        self.message = message
        self.body = body

    def __bytes__(self) -> bytes:
        if "Compress" in self.headers:
            body = zlib.compress(self.body)
        else:
            body = self.body

        if len(body) > 0:
            self.headers["Content-length"] = ContentLengthValue(length=len(body))

        status = self.status_code
        encoded_headers = b"".join(
            [
                b"%b: %b\r\n" % (key.encode("ascii"), bytes(value))
                for key, value in self.headers.items()
            ]
        )
        message = self.message.encode("ascii")

        return (
            b"SPAMD/%(version)b "
            b"%(status)d "
            b"%(message)b\r\n"
            b"%(headers)b\r\n"
            b"%(body)b"
            % {
                b"version": self.version.encode("ascii"),
                b"status": status,
                b"message": message,
                b"headers": encoded_headers,
                b"body": body,
            }
        )

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return (
            f"<{self.__class__.__module__}.{self.__class__.__qualname__} "
            f"[{self.message}] "
            f"object at {hex(id(self))}>"
        )

    def __eq__(self, other: Any) -> bool:
        try:
            return (
                self.version == other.version
                and self.headers == other.headers
                and self.status_code == other.status_code
                and self.message == other.message
                and self.body == other.body
            )
        except AttributeError:
            return False

    @property
    def status_code(self) -> Union[Status, int]:
        """Status code property getter.

        :return: Value of status code.
        """

        return self._status_code

    @status_code.setter
    def status_code(self, code: Union[Status, int]) -> None:
        """Status code property setter.

        :param code: Status code value to set.
        """

        try:
            self._status_code = Status(code)
        except ValueError:
            self._status_code = code

    @property
    def body(self) -> bytes:
        """Body property getter.

        :return: Value of body.
        """

        return self._body

    @body.setter
    def body(self, value: Union[bytes, SupportsBytes]) -> None:
        """Body property setter.

        :param value: Value to set the body.
        """

        self._body = bytes(value)

    def raise_for_status(self) -> None:
        """Raises an exception if the status code isn't zero.

        :raises ResponseException:
        :raises UsageException:
        :raises DataErrorException:
        :raises NoInputException:
        :raises NoUserException:
        :raises NoHostException:
        :raises UnavailableException:
        :raises InternalSoftwareException:
        :raises OSErrorException:
        :raises OSFileException:
        :raises CantCreateException:
        :raises IOErrorException:
        :raises TemporaryFailureException:
        :raises ProtocolException:
        :raises NoPermissionException:
        :raises ConfigException:
        :raises ServerTimeoutException:
        """

        if self.status_code == 0:
            return
        else:
            status_exception = {
                64: UsageException,
                65: DataErrorException,
                66: NoInputException,
                67: NoUserException,
                68: NoHostException,
                69: UnavailableException,
                70: InternalSoftwareException,
                71: OSErrorException,
                72: OSFileException,
                73: CantCreateException,
                74: IOErrorException,
                75: TemporaryFailureException,
                76: ProtocolException,
                77: NoPermissionException,
                78: ConfigException,
                79: ServerTimeoutException,
            }
            if self.status_code in status_exception:
                raise status_exception[self.status_code](self.message, self)
            else:
                raise ResponseException(self.status_code, self.message, self)

    def to_json(self) -> dict[str, Any]:
        """Converts to JSON serializable object."""

        return {
            "version": self.version,
            "status_code": int(self.status_code),
            "message": self.message,
            "headers": {key: value.to_json() for key, value in self.headers.items()},
            "body": b64encode(self.body).decode(),
        }


class ResponseException(Exception):
    """Base class for exceptions raised from a response."""

    def __init__(self, code: int, message: str, response: Response):
        """ResponseException constructor.

        :param code: Response code number.
        :param message: Message response.
        """

        self.code = code
        self.response = response
        super().__init__(message)


class UsageException(ResponseException):
    """Command line usage error."""

    def __init__(self, message: str, response: Response):
        """UsageException constructor.

        :param message: Message response.
        """

        super().__init__(64, message, response)


class DataErrorException(ResponseException):
    """Data format error."""

    def __init__(self, message: str, response: Response):
        """DataErrorException constructor.

        :param message: Message response.
        """

        super().__init__(65, message, response)


class NoInputException(ResponseException):
    """Cannot open input."""

    def __init__(self, message: str, response: Response):
        """NoInputException constructor.

        :param message: Message response.
        """

        super().__init__(66, message, response)


class NoUserException(ResponseException):
    """Addressee unknown."""

    def __init__(self, message: str, response: Response):
        """NoUserException constructor.

        :param message: Message response.
        """

        super().__init__(67, message, response)


class NoHostException(ResponseException):
    """Hostname unknown."""

    def __init__(self, message: str, response: Response):
        """NoHostException constructor.

        :param message: Message response.
        """

        super().__init__(68, message, response)


class UnavailableException(ResponseException):
    """Service unavailable."""

    def __init__(self, message: str, response: Response):
        """UnavailableException constructor.

        :param message: Message response.
        """

        super().__init__(69, message, response)


class InternalSoftwareException(ResponseException):
    """Internal software error."""

    def __init__(self, message: str, response: Response):
        """InternalSoftwareException constructor.

        :param message: Message response.
        """

        super().__init__(70, message, response)


class OSErrorException(ResponseException):
    """System error (e.g. can't fork the process)."""

    def __init__(self, message: str, response: Response):
        """OSErrorException constructor.

        :param message: Message response.
        """

        super().__init__(71, message, response)


class OSFileException(ResponseException):
    """Critical operating system file missing."""

    def __init__(self, message: str, response: Response):
        """OSFileException constructor.

        :param message: Message response.
        """

        super().__init__(72, message, response)


class CantCreateException(ResponseException):
    """Can't create (user) output file."""

    def __init__(self, message: str, response: Response):
        """CantCreateException constructor.

        :param message: Message response.
        """

        super().__init__(73, message, response)


class IOErrorException(ResponseException):
    """Input/output error."""

    def __init__(self, message: str, response: Response):
        """IOErrorException constructor.

        :param message: Message response.
        """

        super().__init__(74, message, response)


class TemporaryFailureException(ResponseException):
    """Temporary failure, user is invited to try again."""

    def __init__(self, message: str, response: Response):
        """TemporaryFailureException constructor.

        :param message: Message response.
        """

        super().__init__(75, message, response)


class ProtocolException(ResponseException):
    """Remote error in protocol."""

    def __init__(self, message: str, response: Response):
        """ProtocolException constructor.

        :param message: Message response.
        """

        super().__init__(76, message, response)


class NoPermissionException(ResponseException):
    """Permission denied."""

    def __init__(self, message: str, response: Response):
        """NoPermissionException constructor.

        :param message: Message response.
        """

        super().__init__(77, message, response)


class ConfigException(ResponseException):
    """Configuration error."""

    def __init__(self, message: str, response: Response):
        """ConfigException constructor.

        :param message: Message response.
        """

        super().__init__(78, message, response)


class ServerTimeoutException(ResponseException, TimeoutException):
    """Timeout exception from the server."""

    def __init__(self, message: str, response: Response):
        """ServerTimeoutException constructor.

        :param message: Message response.
        """

        super().__init__(79, message, response)



"""Frontend functions for the package."""

from __future__ import annotations

import ssl
from functools import partial
from pathlib import Path
from typing import Any, Optional, SupportsBytes, Union, cast

from loguru import logger

from .client import Client
from .connections import ConnectionManagerBuilder, SSLContextBuilder, Timeout
from .header_values import ActionOption, MessageClassOption, MessageClassValue
from .incremental_parser import parse_set_remove_value
from .requests import Request
from .responses import Response


class FrontendClientBuilder:
    """Builds the :class:`aiospamc.client.Client` based off of frontend arguments."""

    def __init__(self):
        """Constructor for FrontendClientBuilder."""

        self._connection_builder = ConnectionManagerBuilder()
        self._ssl = False
        self._ssl_builder = SSLContextBuilder()

    def build(self) -> Client:
        """Builds the :class:`aiospamc.client.Client`.

        :return: An instance of :class:`aiospamc.client.Client`.
        """

        if self._ssl:
            self._connection_builder.add_ssl_context(self._ssl_builder.build())
        connection_manager = self._connection_builder.build()

        return Client(connection_manager)

    def with_connection(
        self,
        host: str = "localhost",
        port: int = 783,
        socket_path: Optional[Path] = None,
    ) -> FrontendClientBuilder:
        """Sets the type of connection manager to use.

        Defaults to TCP, but if a Unix socket is provided it will be used.

        :param host: TCP hostname to use.
        :param port: TCP port to use.
        :param socket_path: Path to the Unix socket.

        :return: This builder instance.
        """

        if socket_path:
            self._connection_builder = self._connection_builder.with_unix_socket(
                socket_path
            )
        else:
            self._connection_builder = self._connection_builder.with_tcp(host, port)

        return self

    def add_verify(
        self, verify: Union[bool, Path, ssl.SSLContext, None] = None
    ) -> FrontendClientBuilder:
        """Adds an SSL context to the connection manager.

        :param verify: How to configure the SSL context. If `True`, add the default
            certificate authorities. If `False`, accept any certificate. If a :class:`pathlib.Path`,
            add the certificates from it. If an :class:`ssl.SSLContext`, then use it.

        :return: This builder instance.
        """

        if verify is None:
            return self

        self._ssl = True

        if verify is True:
            self._ssl_builder.add_default_ca()
        elif verify is False:
            self._ssl_builder.add_default_ca().dont_verify()
        elif isinstance(verify, ssl.SSLContext):
            self._ssl_builder.with_context(verify)
        else:
            self._ssl_builder.add_ca(verify)

        return self

    def add_client_cert(
        self,
        cert: Optional[
            Union[
                Path,
                tuple[Path, Optional[Path]],
                tuple[Path, Optional[Path], Optional[str]],
            ]
        ],
    ) -> FrontendClientBuilder:
        """Add a client certificate to authenticate to the server.

        :param cert: Client certificate. Takes up to three a three tuple value.
            1. Path to the certificate and key.
            2. Path to the certificate and path to the key.
            3. Path to the certificate, path to the key, and password of the key.

        :return: This builder instance.
        """

        if cert is None:
            return self

        if not self._ssl:
            self.add_verify(True)

        def pwd_check(password: Optional[str] = None) -> str:
            """Return the password, otherwise throw an exception.

            :return: The password.
            :raises ValueError: When the password is `None`.
            """

            if password is None:
                raise ValueError("Private key password not provided")
            return password

        if isinstance(cert, Path):
            self._ssl_builder.add_client(cert, password=partial(pwd_check, None))
        elif isinstance(cert, tuple) and len(cert) == 2:
            client, key = cast(tuple[Path, Optional[Path]], cert)
            self._ssl_builder.add_client(client, key, password=partial(pwd_check, None))
        elif isinstance(cert, tuple) and len(cert) == 3:
            client, key, password = cast(
                tuple[Path, Optional[Path], Optional[str]], cert
            )
            self._ssl_builder.add_client(
                client, key, password=partial(pwd_check, password)
            )
        else:
            raise TypeError("Unexepected value")

        return self

    def set_timeout(self, timeout: Optional[Timeout] = None) -> FrontendClientBuilder:
        """Sets the timeout for the connection.

        :param timeout: Timeout object.

        :return: This builder instance.
        """

        if timeout:
            self._connection_builder.set_timeout(timeout)

        return self


def _add_compress_header(request: Request, compress: bool):
    """Adds a compress header to the request if specified.

    :param request: The request to be modified.
    :param compress: Switch if to add compress header.
    """

    if compress:
        request.headers.compress = "zlib"


def _add_user_header(request: Request, user: Optional[str]):
    """Adds a user header to the request if specified.

    :param request: The request to be modified.
    :param user: Optional username to be added.
    """

    if user:
        request.headers.user = user


async def check(
    message: Union[bytes, SupportsBytes],
    *,
    host: str = "localhost",
    port: int = 783,
    socket_path: Optional[Path] = None,
    timeout: Optional[Timeout] = None,
    verify: Union[bool, Path, ssl.SSLContext, None] = None,
    cert: Optional[
        Union[
            Path,
            tuple[Path, Optional[Path]],
            tuple[Path, Optional[Path], Optional[str]],
        ]
    ] = None,
    user: Optional[str] = None,
    compress: bool = False,
) -> Response:
    """Checks a message if it's spam and return a response with a score header.

    :param message: Copy of the message.
    :param host: Hostname or IP address of the SPAMD service, defaults to localhost.
    :param port: Port number for the SPAMD service, defaults to 783.
    :param socket_path: Path to Unix socket.
    :param timeout: Timeout settings.
    :param verify:
        Enable SSL. `True` will use the root certificates from the :py:mod:`certifi` package.
        `False` will use SSL, but not verify the root certificates. Passing a string to a filename
        will use the path to verify the root certificates.
    :param cert: Use client certificate. Can either by a :class:`pathlib.Path` to a file that includes both the
        certificate and key. Can be a tuple containing a :class:`pathlib.Path` to the certificate and a :class:`pathlib.Path` to
        the key. Can be a tuple containing a :class:`pathlib.Path` to the certificate, :class:`pathlib.Path` to the key, and a
        password.
    :param user: Username to pass to the SPAMD service.
    :param compress: Enable compress of the request body.

    :return:
        A successful response with a "Spam" header showing if the message is
        considered spam as well as the score.

    :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
    :raises AIOSpamcConnectionFailed: Raised if an error occurred when trying to connect.
    :raises UsageException: Error in command line usage.
    :raises DataErrorException: Error with data format.
    :raises NoInputException: Cannot open input.
    :raises NoUserException: Addressee unknown.
    :raises NoHostException: Hostname unknown.
    :raises UnavailableException: Service unavailable.
    :raises InternalSoftwareException: Internal software error.
    :raises OSErrorException: System error.
    :raises OSFileException: Operating system file missing.
    :raises CantCreateException: Cannot create output file.
    :raises IOErrorException: Input/output error.
    :raises TemporaryFailureException: Temporary failure, may reattempt.
    :raises ProtocolException: Error in the protocol.
    :raises NoPermissionException: Permission denied.
    :raises ConfigException: Error in configuration.
    :raises ServerTimeoutException: Server returned a response that it timed out.
    :raises ClientTimeoutException: Client timed out during connection.
    """

    req = Request("CHECK", body=bytes(message))
    _add_compress_header(req, compress)
    _add_user_header(req, user)
    context_logger = logger.bind(
        host=host,
        port=port,
        socket_path=socket_path,
        user=user,
        request=req,
    )
    context_logger.info("Sending CHECK request")

    client_builder = FrontendClientBuilder()
    client = (
        client_builder.with_connection(host, port, socket_path)
        .add_verify(verify)
        .add_client_cert(cert)
        .set_timeout(timeout)
        .build()
    )
    try:
        response = await client.request(req)
    except Exception:
        context_logger.exception("Exception when calling check function")
        raise

    context_logger.bind(response=response).success(
        "Successfully completed check function"
    )

    return response


async def headers(
    message: Union[bytes, SupportsBytes],
    *,
    host: str = "localhost",
    port: int = 783,
    socket_path: Optional[Path] = None,
    timeout: Optional[Timeout] = None,
    verify: Union[bool, Path, ssl.SSLContext, None] = None,
    cert: Union[
        Path,
        tuple[Path, Optional[Path]],
        tuple[Path, Optional[Path], Optional[str]],
        None,
    ] = None,
    user: Optional[str] = None,
    compress: bool = False,
) -> Response:
    """Checks a message if it's spam and return the modified message headers.

    :param message: Copy of the message.
    :param host: Hostname or IP address of the SPAMD service, defaults to localhost.
    :param port: Port number for the SPAMD service, defaults to 783.
    :param socket_path: Path to Unix socket.
    :param timeout: Timeout settings.
    :param verify:
        Enable SSL. `True` will use the root certificates from the :py:mod:`certifi` package.
        `False` will use SSL, but not verify the root certificates. Passing a string to a filename
        will use the path to verify the root certificates.
    :param cert: Use client certificate. Can either by a :class:`pathlib.Path` to a file that includes both the
        certificate and key. Can be a tuple containing a :class:`pathlib.Path` to the certificate and a :class:`pathlib.Path` to
        the key. Can be a tuple containing a :class:`pathlib.Path` to the certificate, :class:`pathlib.Path` to the key, and a
        password.
    :param user: Username to pass to the SPAMD service.
    :param compress: Enable compress of the request body.

    :return:
        A successful response with a "Spam" header showing if the message is
        considered spam as well as the score.  The body contains the modified
        message headers, but not the content of the message.

    :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
    :raises AIOSpamcConnectionFailed: Raised if an error occurred when trying to connect.
    :raises UsageException: Error in command line usage.
    :raises DataErrorException: Error with data format.
    :raises NoInputException: Cannot open input.
    :raises NoUserException: Addressee unknown.
    :raises NoHostException: Hostname unknown.
    :raises UnavailableException: Service unavailable.
    :raises InternalSoftwareException: Internal software error.
    :raises OSErrorException: System error.
    :raises OSFileException: Operating system file missing.
    :raises CantCreateException: Cannot create output file.
    :raises IOErrorException: Input/output error.
    :raises TemporaryFailureException: Temporary failure, may reattempt.
    :raises ProtocolException: Error in the protocol.
    :raises NoPermissionException: Permission denied.
    :raises ConfigException: Error in configuration.
    :raises ServerTimeoutException: Server returned a response that it timed out.
    :raises ClientTimeoutException: Client timed out during connection.
    """

    req = Request("HEADERS", body=bytes(message))
    _add_compress_header(req, compress)
    _add_user_header(req, user)
    context_logger = logger.bind(
        host=host,
        port=port,
        socket_path=socket_path,
        user=user,
        request=req,
    )
    context_logger.info("Sending HEADERS request")

    client_builder = FrontendClientBuilder()
    client = (
        client_builder.with_connection(host, port, socket_path)
        .add_verify(verify)
        .add_client_cert(cert)
        .set_timeout(timeout)
        .build()
    )
    try:
        response = await client.request(req)
    except Exception:
        context_logger.exception("Exception when calling headers function")
        raise

    context_logger.bind(response=response).success(
        "Successfully completed headers function"
    )

    return response


async def ping(
    *,
    host: str = "localhost",
    port: int = 783,
    socket_path: Optional[Path] = None,
    timeout: Optional[Timeout] = None,
    verify: Union[bool, Path, ssl.SSLContext, None] = None,
    cert: Union[
        Path,
        tuple[Path, Optional[Path]],
        tuple[Path, Optional[Path], Optional[str]],
        None,
    ] = None,
) -> Response:
    """Sends a ping to the SPAMD service.

    :param host: Hostname or IP address of the SPAMD service, defaults to localhost.
    :param port: Port number for the SPAMD service, defaults to 783.
    :param socket_path: Path to Unix socket.
    :param timeout: Timeout settings.
    :param verify:
        Enable SSL. `True` will use the root certificates from the :py:mod:`certifi` package.
        `False` will use SSL, but not verify the root certificates. Passing a string to a filename
        will use the path to verify the root certificates.
    :param cert: Use client certificate. Can either by a :class:`pathlib.Path` to a file that includes both the
        certificate and key. Can be a tuple containing a :class:`pathlib.Path` to the certificate and a :class:`pathlib.Path` to
        the key. Can be a tuple containing a :class:`pathlib.Path` to the certificate, :class:`pathlib.Path` to the key, and a
        password.

    :return: A response with "PONG".

    :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
    :raises AIOSpamcConnectionFailed: Raised if an error occurred when trying to connect.
    :raises UsageException: Error in command line usage.
    :raises DataErrorException: Error with data format.
    :raises NoInputException: Cannot open input.
    :raises NoUserException: Addressee unknown.
    :raises NoHostException: Hostname unknown.
    :raises UnavailableException: Service unavailable.
    :raises InternalSoftwareException: Internal software error.
    :raises OSErrorException: System error.
    :raises OSFileException: Operating system file missing.
    :raises CantCreateException: Cannot create output file.
    :raises IOErrorException: Input/output error.
    :raises TemporaryFailureException: Temporary failure, may reattempt.
    :raises ProtocolException: Error in the protocol.
    :raises NoPermissionException: Permission denied.
    :raises ConfigException: Error in configuration.
    :raises ServerTimeoutException: Server returned a response that it timed out.
    :raises ClientTimeoutException: Client timed out during connection.
    """

    req = Request("PING")
    context_logger = logger.bind(
        host=host,
        port=port,
        socket_path=socket_path,
        request=req,
    )
    context_logger.info("Sending PING request")

    client_builder = FrontendClientBuilder()
    client = (
        client_builder.with_connection(host, port, socket_path)
        .add_verify(verify)
        .add_client_cert(cert)
        .set_timeout(timeout)
        .build()
    )
    try:
        response = await client.request(req)
    except Exception:
        context_logger.exception("Exception when calling ping function")
        raise

    context_logger.bind(response=response).success(
        "Successfully completed ping function"
    )

    return response


async def process(
    message: Union[bytes, SupportsBytes],
    *,
    host: str = "localhost",
    port: int = 783,
    socket_path: Optional[Path] = None,
    timeout: Optional[Timeout] = None,
    verify: Union[bool, Path, ssl.SSLContext, None] = None,
    cert: Union[
        Path,
        tuple[Path, Optional[Path]],
        tuple[Path, Optional[Path], Optional[str]],
        None,
    ] = None,
    user: Optional[str] = None,
    compress: bool = False,
) -> Response:
    """Checks a message if it's spam and return a response with a score header.

    :param message: Copy of the message.
    :param host: Hostname or IP address of the SPAMD service, defaults to localhost.
    :param port: Port number for the SPAMD service, defaults to 783.
    :param socket_path: Path to Unix socket.
    :param timeout: Timeout settings.
    :param verify:
        Enable SSL. `True` will use the root certificates from the :py:mod:`certifi` package.
        `False` will use SSL, but not verify the root certificates. Passing a string to a filename
        will use the path to verify the root certificates.
    :param cert: Use client certificate. Can either by a :class:`pathlib.Path` to a file that includes both the
        certificate and key. Can be a tuple containing a :class:`pathlib.Path` to the certificate and a :class:`pathlib.Path` to
        the key. Can be a tuple containing a :class:`pathlib.Path` to the certificate, :class:`pathlib.Path` to the key, and a
        password.
    :param user: Username to pass to the SPAMD service.
    :param compress: Enable compress of the request body.

    :return:
        A successful response with a "Spam" header showing if the message is
        considered spam as well as the score.  The body contains a modified
        copy of the message.

    :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
    :raises AIOSpamcConnectionFailed: Raised if an error occurred when trying to connect.
    :raises UsageException: Error in command line usage.
    :raises DataErrorException: Error with data format.
    :raises NoInputException: Cannot open input.
    :raises NoUserException: Addressee unknown.
    :raises NoHostException: Hostname unknown.
    :raises UnavailableException: Service unavailable.
    :raises InternalSoftwareException: Internal software error.
    :raises OSErrorException: System error.
    :raises OSFileException: Operating system file missing.
    :raises CantCreateException: Cannot create output file.
    :raises IOErrorException: Input/output error.
    :raises TemporaryFailureException: Temporary failure, may reattempt.
    :raises ProtocolException: Error in the protocol.
    :raises NoPermissionException: Permission denied.
    :raises ConfigException: Error in configuration.
    :raises ServerTimeoutException: Server returned a response that it timed out.
    :raises ClientTimeoutException: Client timed out during connection.
    """

    req = Request("PROCESS", body=bytes(message))
    _add_compress_header(req, compress)
    _add_user_header(req, user)
    context_logger = logger.bind(
        host=host,
        port=port,
        socket_path=socket_path,
        user=user,
        request=req,
    )
    context_logger.info("Sending PROCESS request")

    client_builder = FrontendClientBuilder()
    client = (
        client_builder.with_connection(host, port, socket_path)
        .add_verify(verify)
        .add_client_cert(cert)
        .set_timeout(timeout)
        .build()
    )
    try:
        response = await client.request(req)
    except Exception:
        context_logger.exception("Exception when calling process function")
        raise

    context_logger.bind(response=response).success(
        "Successfully completed process function"
    )

    return response


async def report(
    message: Union[bytes, SupportsBytes],
    *,
    host: str = "localhost",
    port: int = 783,
    socket_path: Optional[Path] = None,
    timeout: Optional[Timeout] = None,
    verify: Union[bool, Path, ssl.SSLContext, None] = None,
    cert: Union[
        Path,
        tuple[Path, Optional[Path]],
        tuple[Path, Optional[Path], Optional[str]],
        None,
    ] = None,
    user: Optional[str] = None,
    compress: bool = False,
) -> Response:
    """Checks a message if it's spam and return a response with a score header.

    :param message: Copy of the message.
    :param host: Hostname or IP address of the SPAMD service, defaults to localhost.
    :param port: Port number for the SPAMD service, defaults to 783.
    :param socket_path: Path to Unix socket.
    :param timeout: Timeout settings.
    :param verify:
        Enable SSL. `True` will use the root certificates from the :py:mod:`certifi` package.
        `False` will use SSL, but not verify the root certificates. Passing a string to a filename
        will use the path to verify the root certificates.
    :param cert: Use client certificate. Can either by a :class:`pathlib.Path` to a file that includes both the
        certificate and key. Can be a tuple containing a :class:`pathlib.Path` to the certificate and a :class:`pathlib.Path` to
        the key. Can be a tuple containing a :class:`pathlib.Path` to the certificate, :class:`pathlib.Path` to the key, and a
        password.
    :param user: Username to pass to the SPAMD service.
    :param compress: Enable compress of the request body.

    :return:
        A successful response with a "Spam" header showing if the message is
        considered spam as well as the score.  The body contains a report.

    :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
    :raises AIOSpamcConnectionFailed: Raised if an error occurred when trying to connect.
    :raises UsageException: Error in command line usage.
    :raises DataErrorException: Error with data format.
    :raises NoInputException: Cannot open input.
    :raises NoUserException: Addressee unknown.
    :raises NoHostException: Hostname unknown.
    :raises UnavailableException: Service unavailable.
    :raises InternalSoftwareException: Internal software error.
    :raises OSErrorException: System error.
    :raises OSFileException: Operating system file missing.
    :raises CantCreateException: Cannot create output file.
    :raises IOErrorException: Input/output error.
    :raises TemporaryFailureException: Temporary failure, may reattempt.
    :raises ProtocolException: Error in the protocol.
    :raises NoPermissionException: Permission denied.
    :raises ConfigException: Error in configuration.
    :raises ServerTimeoutException: Server returned a response that it timed out.
    :raises ClientTimeoutException: Client timed out during connection.
    """

    req = Request("REPORT", body=bytes(message))
    _add_compress_header(req, compress)
    _add_user_header(req, user)
    context_logger = logger.bind(
        host=host,
        port=port,
        socket_path=socket_path,
        user=user,
        request=req,
    )
    context_logger.info("Sending REPORT request")

    client_builder = FrontendClientBuilder()
    client = (
        client_builder.with_connection(host, port, socket_path)
        .add_verify(verify)
        .add_client_cert(cert)
        .set_timeout(timeout)
        .build()
    )
    try:
        response = await client.request(req)
    except Exception:
        context_logger.exception("Exception when calling report function")
        raise

    context_logger.bind(response=response).success(
        "Successfully completed report function"
    )

    return response


async def report_if_spam(
    message: Union[bytes, SupportsBytes],
    *,
    host: str = "localhost",
    port: int = 783,
    socket_path: Optional[Path] = None,
    timeout: Optional[Timeout] = None,
    verify: Union[bool, Path, ssl.SSLContext, None] = None,
    cert: Union[
        Path,
        tuple[Path, Optional[Path]],
        tuple[Path, Optional[Path], Optional[str]],
        None,
    ] = None,
    user: Optional[str] = None,
    compress: bool = False,
) -> Response:
    """Checks a message if it's spam and return a response with a score header.

    :param message: Copy of the message.
    :param host: Hostname or IP address of the SPAMD service, defaults to localhost.
    :param port: Port number for the SPAMD service, defaults to 783.
    :param socket_path: Path to Unix socket.
    :param timeout: Timeout settings.
    :param verify:
        Enable SSL. `True` will use the root certificates from the :py:mod:`certifi` package.
        `False` will use SSL, but not verify the root certificates. Passing a string to a filename
        will use the path to verify the root certificates.
    :param cert: Use client certificate. Can either by a :class:`pathlib.Path` to a file that includes both the
        certificate and key. Can be a tuple containing a :class:`pathlib.Path` to the certificate and a :class:`pathlib.Path` to
        the key. Can be a tuple containing a :class:`pathlib.Path` to the certificate, :class:`pathlib.Path` to the key, and a
        password.
    :param user: Username to pass to the SPAMD service.
    :param compress: Enable compress of the request body.

    :return:
        A successful response with a "Spam" header showing if the message is
        considered spam as well as the score.  The body contains a report if
        the message is considered spam.

    :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
    :raises AIOSpamcConnectionFailed: Raised if an error occurred when trying to connect.
    :raises UsageException: Error in command line usage.
    :raises DataErrorException: Error with data format.
    :raises NoInputException: Cannot open input.
    :raises NoUserException: Addressee unknown.
    :raises NoHostException: Hostname unknown.
    :raises UnavailableException: Service unavailable.
    :raises InternalSoftwareException: Internal software error.
    :raises OSErrorException: System error.
    :raises OSFileException: Operating system file missing.
    :raises CantCreateException: Cannot create output file.
    :raises IOErrorException: Input/output error.
    :raises TemporaryFailureException: Temporary failure, may reattempt.
    :raises ProtocolException: Error in the protocol.
    :raises NoPermissionException: Permission denied.
    :raises ConfigException: Error in configuration.
    :raises ServerTimeoutException: Server returned a response that it timed out.
    :raises ClientTimeoutException: Client timed out during connection.
    """

    req = Request("REPORT_IFSPAM", body=bytes(message))
    _add_compress_header(req, compress)
    _add_user_header(req, user)
    context_logger = logger.bind(
        host=host,
        port=port,
        socket_path=socket_path,
        user=user,
        request=req,
    )
    context_logger.info("Sending REPORT_IFSPAM request")

    client_builder = FrontendClientBuilder()
    client = (
        client_builder.with_connection(host, port, socket_path)
        .add_verify(verify)
        .add_client_cert(cert)
        .set_timeout(timeout)
        .build()
    )
    try:
        response = await client.request(req)
    except Exception:
        context_logger.exception("Exception when calling report_if_spam function")
        raise

    context_logger.bind(response=response).success(
        "Successfully completed report_if_spam function"
    )

    return response


async def symbols(
    message: Union[bytes, SupportsBytes],
    *,
    host: str = "localhost",
    port: int = 783,
    socket_path: Optional[Path] = None,
    timeout: Optional[Timeout] = None,
    verify: Union[bool, Path, ssl.SSLContext, None] = None,
    cert: Union[
        Path,
        tuple[Path, Optional[Path]],
        tuple[Path, Optional[Path], Optional[str]],
        None,
    ] = None,
    user: Optional[str] = None,
    compress: bool = False,
) -> Response:
    """Checks a message if it's spam and return a response with rules that matched.

    :param message: Copy of the message.
    :param host: Hostname or IP address of the SPAMD service, defaults to localhost.
    :param port: Port number for the SPAMD service, defaults to 783.
    :param socket_path: Path to Unix socket.
    :param timeout: Timeout settings.
    :param verify:
        Enable SSL. `True` will use the root certificates from the :py:mod:`certifi` package.
        `False` will use SSL, but not verify the root certificates. Passing a string to a filename
        will use the path to verify the root certificates.
    :param cert: Use client certificate. Can either by a :class:`pathlib.Path` to a file that includes both the
        certificate and key. Can be a tuple containing a :class:`pathlib.Path` to the certificate and a :class:`pathlib.Path` to
        the key. Can be a tuple containing a :class:`pathlib.Path` to the certificate, :class:`pathlib.Path` to the key, and a
        password.
    :param user: Username to pass to the SPAMD service.
    :param compress: Enable compress of the request body.

    :return:
        A successful response with a "Spam" header showing if the message is
        considered spam as well as the score.  The body contains a
        comma-separated list of the symbols that were hit.

    :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
    :raises AIOSpamcConnectionFailed: Raised if an error occurred when trying to connect.
    :raises UsageException: Error in command line usage.
    :raises DataErrorException: Error with data format.
    :raises NoInputException: Cannot open input.
    :raises NoUserException: Addressee unknown.
    :raises NoHostException: Hostname unknown.
    :raises UnavailableException: Service unavailable.
    :raises InternalSoftwareException: Internal software error.
    :raises OSErrorException: System error.
    :raises OSFileException: Operating system file missing.
    :raises CantCreateException: Cannot create output file.
    :raises IOErrorException: Input/output error.
    :raises TemporaryFailureException: Temporary failure, may reattempt.
    :raises ProtocolException: Error in the protocol.
    :raises NoPermissionException: Permission denied.
    :raises ConfigException: Error in configuration.
    :raises ServerTimeoutException: Server returned a response that it timed out.
    :raises ClientTimeoutException: Client timed out during connection.
    """

    req = Request("SYMBOLS", body=bytes(message))
    _add_compress_header(req, compress)
    _add_user_header(req, user)
    context_logger = logger.bind(
        host=host,
        port=port,
        socket_path=socket_path,
        user=user,
        request=req,
    )
    context_logger.info("Sending SYMBOLS request")

    client_builder = FrontendClientBuilder()
    client = (
        client_builder.with_connection(host, port, socket_path)
        .add_verify(verify)
        .add_client_cert(cert)
        .set_timeout(timeout)
        .build()
    )
    try:
        response = await client.request(req)
    except Exception:
        context_logger.exception("Exception when calling symbols function")
        raise

    context_logger.bind(response=response).success(
        "Successfully completed symbols function"
    )

    return response


async def tell(
    message: Union[bytes, SupportsBytes],
    message_class: Union[str, MessageClassOption],
    remove_action: Union[str, ActionOption, None] = None,
    set_action: Union[str, ActionOption, None] = None,
    *,
    host: str = "localhost",
    port: int = 783,
    socket_path: Optional[Path] = None,
    timeout: Optional[Timeout] = None,
    verify: Optional[Union[bool, Path, ssl.SSLContext]] = None,
    cert: Union[
        Path,
        tuple[Path, Optional[Path]],
        tuple[Path, Optional[Path], Optional[str]],
        None,
    ] = None,
    user: Optional[str] = None,
    compress: bool = False,
) -> Response:
    """Checks a message if it's spam and return a response with a score header.

    :param message: Copy of the message.
    :param message_class: Classify the message as 'spam' or 'ham'.
    :param remove_action: Remove message class for message in database.
    :param set_action: Set message class for message in database.
    :param host: Hostname or IP address of the SPAMD service, defaults to localhost.
    :param port: Port number for the SPAMD service, defaults to 783.
    :param socket_path: Path to Unix socket.
    :param timeout: Timeout settings.
    :param verify:
        Enable SSL. `True` will use the root certificates from the :py:mod:`certifi` package.
        `False` will use SSL, but not verify the root certificates. Passing a string to a filename
        will use the path to verify the root certificates.
    :param cert: Use client certificate. Can either by a :class:`pathlib.Path` to a file that includes both the
        certificate and key. Can be a tuple containing a :class:`pathlib.Path` to the certificate and a :class:`pathlib.Path` to
        the key. Can be a tuple containing a :class:`pathlib.Path` to the certificate, :class:`pathlib.Path` to the key, and a
        password.
    :param user: Username to pass to the SPAMD service.
    :param compress: Enable compress of the request body.

    :return:
        A successful response with "DidSet" and/or "DidRemove" headers along with the
        actions that were taken.

    :raises BadResponse: If the response from SPAMD is ill-formed this exception will be raised.
    :raises AIOSpamcConnectionFailed: Raised if an error occurred when trying to connect.
    :raises UsageException: Error in command line usage.
    :raises DataErrorException: Error with data format.
    :raises NoInputException: Cannot open input.
    :raises NoUserException: Addressee unknown.
    :raises NoHostException: Hostname unknown.
    :raises UnavailableException: Service unavailable.
    :raises InternalSoftwareException: Internal software error.
    :raises OSErrorException: System error.
    :raises OSFileException: Operating system file missing.
    :raises CantCreateException: Cannot create output file.
    :raises IOErrorException: Input/output error.
    :raises TemporaryFailureException: Temporary failure, may reattempt.
    :raises ProtocolException: Error in the protocol.
    :raises NoPermissionException: Permission denied.
    :raises ConfigException: Error in configuration.
    :raises ServerTimeoutException: Server returned a response that it timed out.
    :raises ClientTimeoutException: Client timed out during connection.
    """

    headers: dict[str, Any] = {
        "Message-class": MessageClassValue(MessageClassOption(message_class))
    }
    if remove_action:
        headers["Remove"] = parse_set_remove_value(remove_action)
    if set_action:
        headers["Set"] = parse_set_remove_value(set_action)
    req = Request("TELL", headers=headers, body=bytes(message))
    _add_compress_header(req, compress)
    _add_user_header(req, user)
    context_logger = logger.bind(
        host=host,
        port=port,
        socket_path=socket_path,
        user=user,
        request=req,
    )
    context_logger.info("Sending TELL request")

    client_builder = FrontendClientBuilder()
    client = (
        client_builder.with_connection(host, port, socket_path)
        .add_verify(verify)
        .add_client_cert(cert)
        .set_timeout(timeout)
        .build()
    )
    try:
        response = await client.request(req)
    except Exception:
        context_logger.exception("Exception when calling tell function")
        raise

    context_logger.bind(response=response).success(
        "Successfully completed tell function"
    )

    return response



"""CLI commands."""

import asyncio
import json
import ssl
import sys
from enum import Enum
from getpass import getpass, getuser
from io import BufferedReader
from pathlib import Path
from typing import Optional

import typer
from loguru import logger
from typing_extensions import Annotated

from aiospamc.exceptions import (
    AIOSpamcConnectionFailed,
    BadResponse,
    ClientTimeoutException,
)
from aiospamc.header_values import (
    ActionOption,
    Headers,
    MessageClassOption,
    MessageClassValue,
    SetOrRemoveValue,
)

from . import __version__
from .client import Client, Request
from .connections import ConnectionManagerBuilder, SSLContextBuilder, Timeout
from .responses import Response, ResponseException

app = typer.Typer(no_args_is_help=True)


class Output(str, Enum):
    """Output formats."""

    Json = "json"
    Text = "text"

    def __str__(self) -> str:
        return self.value


# Exit codes
SUCCESS = NOT_SPAM = PING_SUCCESS = 0
IS_SPAM = REPORT_FAILED = REVOKE_FAILED = 1
BAD_RESPONSE = 3
TIMEOUT_ERROR = 4
CONNECTION_ERROR = 5
UNEXPECTED_ERROR = 6
FILE_NOT_FOUND_ERROR = 7


class CliClientBuilder:
    """Client builder for CLI arguments."""

    def __init__(self):
        """Constructor for the CliClientBuilder."""

        self._connection_builder = ConnectionManagerBuilder()
        self._ssl = False
        self._ssl_builder = SSLContextBuilder()

    def build(self) -> Client:
        """Builds the :class:`aiospamc.client.Client`.

        :return: An instance of :class:`aiospamc.client.Client`.
        """

        if self._ssl:
            ssl_context = self._ssl_builder.build()
            self._connection_builder.add_ssl_context(ssl_context)
        connection_manager = self._connection_builder.build()

        return Client(connection_manager)

    def with_connection(
        self,
        host: str = "localhost",
        port: int = 783,
        socket_path: Optional[Path] = None,
    ) -> "CliClientBuilder":
        """Sets the type of connection manager to use.

        Defaults to TCP, but if a Unix socket is provided it will be used.

        :param host: TCP hostname to use.
        :param port: TCP port to use.
        :param socket_path: Path to the Unix socket.

        :return: This builder instance.
        """

        if socket_path:
            self._connection_builder = self._connection_builder.with_unix_socket(
                socket_path
            )
        else:
            self._connection_builder = self._connection_builder.with_tcp(host, port)

        return self

    def set_timeout(self, timeout: Timeout) -> "CliClientBuilder":
        """Sets the timeout for the connection.

        :param timeout: Timeout object.

        :return: This builder instance.
        """

        self._connection_builder.set_timeout(timeout)

        return self

    def add_verify(self, verify: bool) -> "CliClientBuilder":
        """Adds an SSL context to the connection manager.

        :param verify: How to configure the SSL context. If `True`, add the default
            certificate authorities. If `False`, accept any certificate.

        :return: This builder instance.
        """

        self._ssl = True
        self._ssl_builder.add_default_ca()
        if not verify:
            self._ssl_builder.dont_verify()

        return self

    def add_ca_cert(self, ca_cert: Optional[Path]) -> "CliClientBuilder":
        """Adds trusted certificate authorities.

        :param ca_cert: Path to the cerficiate file or directory.

        :return: This builder instance.
        """

        if ca_cert is None:
            return self

        self._ssl = True
        if ca_cert.is_dir():
            self._ssl_builder.add_ca_dir(ca_cert)
        elif ca_cert.is_file():
            self._ssl_builder.add_ca_file(ca_cert)
        else:
            raise FileNotFoundError(
                f"Unable to find CA certificate file/directory {ca_cert}"
            )

        return self

    def add_client_cert(
        self,
        cert: Optional[Path],
        key: Optional[Path] = None,
        password: Optional[str] = None,
    ) -> "CliClientBuilder":
        """Add a client certificate to authenticate to the server.

        :param cert: Path to the client certificate and optionally the key.
        :param key: Path to the client key.
        :param password: Password of the client key.

        :return: This builder instance.
        """

        if cert is None:
            return self

        if self._ssl is False:
            self._ssl = True
            self.add_verify(True)
        self._ssl_builder.add_client(
            cert, key, lambda: password or getpass("Private key password")
        )

        return self


class CommandRunner:
    """Object to execute requests and handle exceptions."""

    def __init__(
        self,
        client: Client,
        request: Request,
        output: Output = Output.Text,
    ):
        """CommandRunner constructor.

        :param request: Request to send.
        :param response: Response if returned from server.
        :param output: Output format when printing to console.
        """

        self.client = client
        self.request = request
        self.response: Optional[Response] = None
        self.output = output
        self.exception: Optional[Exception] = None
        self.exit_code = SUCCESS

        self._logger = logger.bind(request=request)

    async def run(self) -> Response:
        """Send the request, get the response and handle common exceptions.

        :return: The response.
        """

        self._logger.info("Sending request")

        try:
            response = await self.client.request(self.request)
        except BadResponse as e:
            self._logger.exception(e)
            self.exit_code = BAD_RESPONSE
            self.exception = e
            self.exit("Error parsing response", True)
        except ResponseException as e:
            self._logger = self._logger.bind(response=e.response)
            self._logger.exception(e)
            self.response = e.response
            self.exit_code = int(self.response.status_code)
            self.exception = e
            self.exit(f"Response error from server: {self.response.message}", True)
        except (asyncio.TimeoutError, ClientTimeoutException) as e:
            self._logger.exception(e)
            self.exit_code = TIMEOUT_ERROR
            self.exception = e
            self.exit("Error: timeout", True)
        except (AIOSpamcConnectionFailed, OSError, ConnectionError, ssl.SSLError) as e:
            self._logger.exception(e)
            self.exit_code = CONNECTION_ERROR
            self.exception = e
            self.exit("Error: Connection error", True)

        self._logger = self._logger.bind(response=response)
        self._logger.success("Successfully recieved request")
        self.response = response

        return response

    def to_json(self) -> str:
        """Converts the object to a JSON string.

        :return: JSON string.
        """

        obj = {
            "request": self.request.to_json(),
            "response": self.response.to_json() if self.response is not None else None,
            "exit_code": self.exit_code,
        }

        return json.dumps(obj, indent=4)

    def exit(self, message: str, err=False):
        """Exits the program, echoing the message if outputting text.
        Otherwise prints the JSON object.

        :param message: Message text to print.
        :param err: Flag if message is an error.
        """

        if self.output == Output.Text:
            typer.echo(message, err=err)
        else:
            typer.echo(self.to_json())
        raise typer.Exit(self.exit_code)


@app.command()
def ping(
    host: Annotated[
        str,
        typer.Option(
            "-h",
            "--host",
            metavar="HOSTNAME",
            help="Hostname to use when connecting using TCP",
        ),
    ] = "localhost",
    port: Annotated[
        int,
        typer.Option(
            "-p", "--port", metavar="PORT", help="Port to use when connecting using TCP"
        ),
    ] = 783,
    socket_path: Annotated[
        Optional[Path],
        typer.Option(
            metavar="PATH", help="Path to use when connecting using Unix sockets"
        ),
    ] = None,
    ssl: Annotated[
        bool, typer.Option(help="Use SSL to communicate with the daemon.")
    ] = False,
    timeout: Annotated[
        float, typer.Option(metavar="SECONDS", help="Timeout in seconds")
    ] = 10,
    out: Annotated[Output, typer.Option(help="Output format for stdout")] = Output.Text,
    ca_cert: Annotated[
        Optional[Path],
        typer.Option(
            help="Path to the CA certificates if overriding",
            envvar="AIOSPAMC_CERT_FILE",
        ),
    ] = None,
    client_cert: Annotated[
        Optional[Path], typer.Option(help="Filename of the client certificate")
    ] = None,
    client_key: Annotated[
        Optional[Path], typer.Option(help="Filename of the client certificate's key")
    ] = None,
    key_password: Annotated[
        Optional[str], typer.Option(help="Password for the client certificate key")
    ] = None,
):
    """Pings the SpamAssassin daemon.

    A successful pong exits with code 0.
    """

    client_builder = CliClientBuilder()
    client_builder.with_connection(host, port, socket_path).set_timeout(
        Timeout(timeout)
    )
    if ssl:
        client_builder.add_verify(True).add_ca_cert(ca_cert).add_client_cert(
            client_cert, client_key, key_password
        )
    client = client_builder.build()

    request = Request("PING")
    runner = CommandRunner(client, request, out)
    response = asyncio.run(runner.run())
    runner.exit(response.message)


def read_message(file: Optional[BufferedReader]) -> bytes:
    """Utility function to read data from stdin.

    :param file: File-like object.
    """

    if not file:
        return sys.stdin.buffer.read()
    return file.read()


@app.command()
def check(
    message: Annotated[
        Optional[typer.FileBinaryRead],
        typer.Argument(show_default=False, help="Filename of message"),
    ] = None,
    host: Annotated[
        str,
        typer.Option(
            "-h",
            "--host",
            metavar="HOSTNAME",
            help="Hostname to use when connecting using TCP",
        ),
    ] = "localhost",
    port: Annotated[
        int,
        typer.Option(
            "-p",
            "--port",
            metavar="PORT",
            help="Port to use when connecting using TCP",
        ),
    ] = 783,
    socket_path: Annotated[
        Optional[Path],
        typer.Option(
            metavar="PATH", help="Path to use when connecting using Unix sockets"
        ),
    ] = None,
    ssl: Annotated[
        bool, typer.Option(help="Use SSL to communicate with the daemon.")
    ] = False,
    user: Annotated[
        str, typer.Option(metavar="USERNAME", help="User to send the request as.")
    ] = getuser(),
    timeout: Annotated[
        float, typer.Option(metavar="SECONDS", help="Timeout in seconds")
    ] = 10,
    out: Annotated[Output, typer.Option(help="Output format for stdout")] = Output.Text,
    ca_cert: Annotated[
        Optional[Path],
        typer.Option(
            help="Path to the CA certificates if overriding",
            envvar="AIOSPAMC_CERT_FILE",
        ),
    ] = None,
    client_cert: Annotated[
        Optional[Path], typer.Option(help="Filename of the client certificate")
    ] = None,
    client_key: Annotated[
        Optional[Path], typer.Option(help="Filename of the client certificate's key")
    ] = None,
    key_password: Annotated[
        Optional[str], typer.Option(help="Password for the client certificate key")
    ] = None,
):
    """Submits a message to SpamAssassin and returns the processed message."""

    message_data = read_message(message)
    headers = Headers()
    headers.user = user
    request = Request("PROCESS", headers=headers, body=message_data)

    client_builder = CliClientBuilder()
    client_builder.with_connection(host, port, socket_path).set_timeout(
        Timeout(timeout)
    )
    if ssl:
        client_builder.add_verify(True).add_ca_cert(ca_cert).add_client_cert(
            client_cert, client_key, key_password
        )
    client = client_builder.build()

    runner = CommandRunner(client, request, out)
    response = asyncio.run(runner.run())

    if spam_header := response.headers.spam:
        if spam_header.value:
            runner.exit_code = IS_SPAM
        runner.exit(f"{spam_header.score}/{spam_header.threshold}")
    else:
        typer.echo("Could not find 'Spam' header", err=True)
        raise typer.Exit(UNEXPECTED_ERROR)


@app.command()
def learn(
    message: Annotated[
        Optional[typer.FileBinaryRead],
        typer.Argument(help="Filename of message"),
    ] = None,
    message_class: Annotated[
        MessageClassOption, typer.Option(help="Message class to classify the message")
    ] = MessageClassOption.spam,
    host: Annotated[
        str,
        typer.Option(
            "-h",
            "--host",
            metavar="HOSTNAME",
            help="Hostname to use when connecting using TCP",
        ),
    ] = "localhost",
    port: Annotated[
        int,
        typer.Option(
            "-p",
            "--port",
            metavar="PORT",
            help="Port to use when connecting using TCP",
        ),
    ] = 783,
    socket_path: Annotated[
        Optional[Path],
        typer.Option(
            metavar="PATH", help="Path to use when connecting using Unix sockets"
        ),
    ] = None,
    ssl: Annotated[
        bool, typer.Option(help="Use SSL to communicate with the daemon.")
    ] = False,
    user: Annotated[
        str, typer.Option(metavar="USERNAME", help="User to send the request as.")
    ] = getuser(),
    timeout: Annotated[
        float, typer.Option(metavar="SECONDS", help="Timeout in seconds")
    ] = 10,
    out: Annotated[Output, typer.Option(help="Output format for stdout")] = Output.Text,
    ca_cert: Annotated[
        Optional[Path],
        typer.Option(
            help="Path to the CA certificates if overriding",
            envvar="AIOSPAMC_CERT_FILE",
        ),
    ] = None,
    client_cert: Annotated[
        Optional[Path], typer.Option(help="Filename of the client certificate")
    ] = None,
    client_key: Annotated[
        Optional[Path], typer.Option(help="Filename of the client certificate's key")
    ] = None,
    key_password: Annotated[
        Optional[str], typer.Option(help="Password for the client certificate key")
    ] = None,
):
    """Ask server to learn the message as spam or ham."""

    message_data = read_message(message)
    headers = Headers()
    headers.user = user
    headers.message_class = message_class
    headers.set_ = ActionOption(local=True, remote=False)
    request = Request(
        "TELL",
        headers=headers,
        body=message_data,
    )

    client_builder = CliClientBuilder()
    client_builder.with_connection(host, port, socket_path).set_timeout(
        Timeout(timeout)
    )
    if ssl:
        client_builder.add_verify(True).add_ca_cert(ca_cert).add_client_cert(
            client_cert, client_key, key_password
        )
    client = client_builder.build()

    runner = CommandRunner(client, request, out)
    response = asyncio.run(runner.run())

    if response.headers.did_set:
        runner.exit("Message successfully learned")
    else:
        runner.exit("Message was already learned")


@app.command()
def forget(
    message: Annotated[
        Optional[typer.FileBinaryRead],
        typer.Argument(help="Filename of message"),
    ] = None,
    host: Annotated[
        str,
        typer.Option(
            "-h",
            "--host",
            metavar="HOSTNAME",
            help="Hostname to use when connecting using TCP",
        ),
    ] = "localhost",
    port: Annotated[
        int,
        typer.Option(
            "-p",
            "--port",
            metavar="PORT",
            help="Port to use when connecting using TCP",
        ),
    ] = 783,
    socket_path: Annotated[
        Optional[Path],
        typer.Option(
            metavar="PATH", help="Path to use when connecting using Unix sockets"
        ),
    ] = None,
    ssl: Annotated[
        bool, typer.Option(help="Use SSL to communicate with the daemon.")
    ] = False,
    user: Annotated[
        str, typer.Option(metavar="USERNAME", help="User to send the request as.")
    ] = getuser(),
    timeout: Annotated[
        float, typer.Option(metavar="SECONDS", help="Timeout in seconds")
    ] = 10,
    out: Annotated[Output, typer.Option(help="Output format for stdout")] = Output.Text,
    ca_cert: Annotated[
        Optional[Path],
        typer.Option(
            help="Path to the CA certificates if overriding",
            envvar="AIOSPAMC_CERT_FILE",
        ),
    ] = None,
    client_cert: Annotated[
        Optional[Path], typer.Option(help="Filename of the client certificate")
    ] = None,
    client_key: Annotated[
        Optional[Path], typer.Option(help="Filename of the client certificate's key")
    ] = None,
    key_password: Annotated[
        Optional[str], typer.Option(help="Password for the client certificate key")
    ] = None,
):
    """Forgets the classification of a message."""

    message_data = read_message(message)
    headers = Headers()
    headers.user = user
    headers.remove = ActionOption(local=True, remote=False)
    request = Request(
        "TELL",
        headers=headers,
        body=message_data,
    )

    client_builder = CliClientBuilder()
    client_builder.with_connection(host, port, socket_path).set_timeout(
        Timeout(timeout)
    )
    if ssl:
        client_builder.add_verify(True).add_ca_cert(ca_cert).add_client_cert(
            client_cert, client_key, key_password
        )
    client = client_builder.build()

    runner = CommandRunner(client, request, out)
    response = asyncio.run(runner.run())

    if response.headers.did_remove:
        runner.exit("Message successfully forgotten")
    else:
        runner.exit("Message was already forgotten")


@app.command()
def report(
    message: Annotated[
        Optional[typer.FileBinaryRead],
        typer.Argument(help="Filename of message"),
    ] = None,
    host: Annotated[
        str,
        typer.Option(
            "-h",
            "--host",
            metavar="HOSTNAME",
            help="Hostname to use when connecting using TCP",
        ),
    ] = "localhost",
    port: Annotated[
        int,
        typer.Option(
            "-p",
            "--port",
            metavar="PORT",
            help="Port to use when connecting using TCP",
        ),
    ] = 783,
    socket_path: Annotated[
        Optional[Path],
        typer.Option(
            metavar="PATH", help="Path to use when connecting using Unix sockets"
        ),
    ] = None,
    ssl: Annotated[
        bool, typer.Option(help="Use SSL to communicate with the daemon.")
    ] = False,
    user: Annotated[
        str, typer.Option(metavar="USERNAME", help="User to send the request as.")
    ] = getuser(),
    timeout: Annotated[
        float, typer.Option(metavar="SECONDS", help="Timeout in seconds")
    ] = 10,
    out: Annotated[Output, typer.Option(help="Output format for stdout")] = Output.Text,
    ca_cert: Annotated[
        Optional[Path],
        typer.Option(
            help="Path to the CA certificates if overriding",
            envvar="AIOSPAMC_CERT_FILE",
        ),
    ] = None,
    client_cert: Annotated[
        Optional[Path], typer.Option(help="Filename of the client certificate")
    ] = None,
    client_key: Annotated[
        Optional[Path], typer.Option(help="Filename of the client certificate's key")
    ] = None,
    key_password: Annotated[
        Optional[str], typer.Option(help="Password for the client certificate key")
    ] = None,
):
    """Report a message to collaborative filtering databases as spam."""

    message_data = read_message(message)
    headers = Headers()
    headers.user = user
    headers.message_class = MessageClassOption.spam
    request = Request(
        "TELL",
        headers=headers,
        body=message_data,
    )

    client_builder = CliClientBuilder()
    client_builder.with_connection(host, port, socket_path).set_timeout(
        Timeout(timeout)
    )
    if ssl:
        client_builder.add_verify(True).add_ca_cert(ca_cert).add_client_cert(
            client_cert, client_key, key_password
        )
    client = client_builder.build()

    runner = CommandRunner(client, request, out)
    response = asyncio.run(runner.run())

    if response.headers.did_set and response.headers.did_set.remote is True:
        runner.exit("Message successfully reported")
    else:
        runner.exit_code = REPORT_FAILED
        runner.exit("Unable to report message")


@app.command()
def revoke(
    message: Annotated[
        Optional[typer.FileBinaryRead],
        typer.Argument(help="Filename of message"),
    ] = None,
    host: Annotated[
        str,
        typer.Option(
            "-h",
            "--host",
            metavar="HOSTNAME",
            help="Hostname to use when connecting using TCP",
        ),
    ] = "localhost",
    port: Annotated[
        int,
        typer.Option(
            "-p",
            "--port",
            metavar="PORT",
            help="Port to use when connecting using TCP",
        ),
    ] = 783,
    socket_path: Annotated[
        Optional[Path],
        typer.Option(
            metavar="PATH", help="Path to use when connecting using Unix sockets"
        ),
    ] = None,
    ssl: Annotated[
        bool, typer.Option(help="Use SSL to communicate with the daemon.")
    ] = False,
    user: Annotated[
        str, typer.Option(metavar="USERNAME", help="User to send the request as.")
    ] = getuser(),
    timeout: Annotated[
        float, typer.Option(metavar="SECONDS", help="Timeout in seconds")
    ] = 10,
    out: Annotated[Output, typer.Option(help="Output format for stdout")] = Output.Text,
    ca_cert: Annotated[
        Optional[Path],
        typer.Option(
            help="Path to the CA certificates if overriding",
            envvar="AIOSPAMC_CERT_FILE",
        ),
    ] = None,
    client_cert: Annotated[
        Optional[Path], typer.Option(help="Filename of the client certificate")
    ] = None,
    client_key: Annotated[
        Optional[Path], typer.Option(help="Filename of the client certificate's key")
    ] = None,
    key_password: Annotated[
        Optional[str], typer.Option(help="Password for the client certificate key")
    ] = None,
):
    """Revoke a message to collaborative filtering databases."""

    message_data = read_message(message)
    headers = Headers()
    headers.user = user
    headers.message_class = MessageClassOption.ham
    request = Request(
        "TELL",
        headers={
            "Message-class": MessageClassValue(MessageClassOption.ham),
            "Remove": SetOrRemoveValue(ActionOption(local=True, remote=True)),
        },
        body=message_data,
    )

    client_builder = CliClientBuilder()
    client_builder.with_connection(host, port, socket_path).set_timeout(
        Timeout(timeout)
    )
    if ssl:
        client_builder.add_verify(True).add_ca_cert(ca_cert).add_client_cert(
            client_cert, client_key, key_password
        )
    client = client_builder.build()

    runner = CommandRunner(client, request, out)
    response = asyncio.run(runner.run())

    if response.headers.did_remove and response.headers.did_remove.remote:
        runner.exit("Message successfully revoked")
    else:
        runner.exit_code = REVOKE_FAILED
        runner.exit("Unable to revoke message")


def version_callback(version: bool):
    """Callback to print the version.

    :param version: Switch on whether to print and exit.
    """

    if version:
        typer.echo(__version__)
        raise typer.Exit()


def debug_callback(debug: bool):
    """Callback to enable debug logging.

    :param debug: Switch on whether to enable debug logging.
    """

    if debug:
        logger.enable(__package__)


@app.callback()
def main(
    version: Annotated[
        bool,
        typer.Option(
            "--version",
            callback=version_callback,
            is_eager=True,
            help="Print version of aiospamc",
        ),
    ] = False,
    debug: Annotated[
        bool,
        typer.Option(
            "--debug",
            is_flag=True,
            callback=debug_callback,
            help="Enable debug logging",
        ),
    ] = False,
):
    """aiospamc sends messages to the SpamAssasin daemon."""

    pass



"""Module for the parsing functions and objects."""

from __future__ import annotations

import re
from enum import Enum, auto
from typing import Any, Callable, Mapping, Union

import loguru
from loguru import logger

from .exceptions import NotEnoughDataError, ParseError, TooMuchDataError
from .header_values import (
    ActionOption,
    BytesHeaderValue,
    CompressValue,
    ContentLengthValue,
    GenericHeaderValue,
    MessageClassOption,
    MessageClassValue,
    SetOrRemoveValue,
    SpamValue,
    UserValue,
)


class States(Enum):
    """States for the parser state machine."""

    Status = auto()
    Header = auto()
    Body = auto()
    Done = auto()


class Parser:
    """The parser state machine.

    :ivar result: Storage location for parsing results."""

    def __init__(
        self,
        delimiter: bytes,
        status_parser: Callable[[bytes], Mapping[str, str]],
        header_parser: Callable[[bytes], tuple[str, Any]],
        body_parser: Callable[[bytes, int], bytes],
        start: States = States.Status,
    ) -> None:
        """Parser constructor.

        :param delimiter: Byte string to split the different sections of the message.
        :param status_parser: Callable to parse the status line of the message.
        :param header_parser: Callable to parse each header line of the message.
        :param body_parser: Callable to parse the body of the message.
        :param start: The state to start the parser on.  Allowed for easier testing.
        """

        self.delimiter = delimiter
        self.status_parser = status_parser
        self.header_parser = header_parser
        self.body_parser = body_parser
        self.result: dict[str, Any] = {"headers": {}, "body": b""}

        self._state = start
        self.buffer = b""

        self._logger = logger

    def _bind(self, **kwargs) -> loguru.Logger:
        """Helper method to bind a new logger.

        :return: A new logger instance.
        """

        self._logger = self._logger.bind(**kwargs)
        return self._logger

    @property
    def state(self) -> States:
        """The current state of the parser.

        :return: The :class:`States` instance.
        """

        return self._state

    def parse(self, stream: bytes) -> Mapping[str, Any]:
        """Entry method to parse a message.

        :param stream: Byte string to parse.

        :return: Returns the parser results dictionary stored in the class attribute :attr:`result`.

        :raises NotEnoughDataError: Raised when not enough data is sent to be parsed.
        :raises TooMuchDataError: Raised when too much data is sent to be parsed.
        :raises ParseError: Raised when a general parse error is found.
        """

        self.buffer += stream

        while self._state != States.Done:
            if self._state == States.Status:
                self.status()
            elif self._state == States.Header:
                self.header()
            elif self._state == States.Body:  # pragma: no branch
                self.body()

        return self.result

    def status(self) -> None:
        """Splits the message at the delimiter and sends the first part of the message to the :attr:`status_parser` callable to
        be parsed.  If successful then the results are stored in the :attr:`result` class attribute and the state
        transitions to :class:`States.Header`.

        :raises NotEnoughDataError: When there is no delimiter the message is incomplete.
        :raises ParseError: When the :attr:`status_parser` callable experiences an error.
        """

        status_line, delimiter, leftover = self.buffer.partition(self.delimiter)

        if status_line and delimiter:
            self.buffer = leftover
            parsed_status = self.status_parser(status_line)
            self.result = {**self.result, **parsed_status}
            self._state = States.Header
            self._bind(**parsed_status)
            self._logger.debug("Finished parsing status line")
        else:
            raise NotEnoughDataError

    def header(self) -> None:
        """Splits the message at the delimiter and sends the line to the :attr:`header_parser`.

        When splitting the action will be determined depending what is matched:

        +-------------+-----------+-----------+----------------------------------------------------------------------+
        | Header line | Delimiter | Leftover  | Action                                                               |
        +=============+===========+===========+======================================================================+
        |     No      |    Yes    | Delimiter | Headers done, empty body. Clear buffer and transition to             |
        |             |           |           | :class:`States.Body`.                                                |
        +-------------+-----------+-----------+----------------------------------------------------------------------+
        |     No      |    Yes    |    N/A    | Headers done.  Transition to :class:`States.Body`.                   |
        +-------------+-----------+-----------+----------------------------------------------------------------------+
        |     Yes     |    Yes    |    N/A    | Parse header.  Record in :attr:`result` class attribute.             |
        +-------------+-----------+-----------+----------------------------------------------------------------------+
        |     No      |    No     |    No     | Message was a status line only.  Transition to :class:`States.Body`. |
        +-------------+-----------+-----------+----------------------------------------------------------------------+

        :raises ParseError: None of the previous conditions are matched.
        """

        header_line, delimiter, leftover = self.buffer.partition(self.delimiter)

        if self._at_end_of_headers_with_empty_body(header_line, delimiter, leftover):
            self.buffer = b""
            self._state = States.Body
            self._bind(headers=self.result["headers"])
            self._logger.debug("Finished parsing headers")
        elif self._at_end_of_headers(header_line, delimiter):
            self.buffer = leftover
            self._state = States.Body
            self._bind(headers=self.result["headers"])
            self._logger.debug("Finished parsing headers")
        elif self._at_next_header(header_line, delimiter):
            self.buffer = leftover
            key, value = self.header_parser(header_line)
            self.result["headers"][key] = value
            self._logger.debug("Parsed header {}", key)
        elif self._is_status_line_only(header_line, delimiter, leftover):
            self._state = States.Body
            self._logger.debug("No headers to parse")
        else:
            raise ParseError("Header section not in recognizable format")

    def _at_end_of_headers_with_empty_body(
        self, header_line: bytes, delimiter: bytes, leftover: bytes
    ) -> bool:
        """Helper method to check if the header sections is done and there is an empty body.

        :param header_line: Contents of the header line.
        :param delimiter: Separator between the header line and the rest of the response.
        :param leftover: Remaining response contents.

        :return: If the header section is finished, and there is no remaining content for a body.
        """

        return all([not header_line, delimiter, leftover == self.delimiter])

    @staticmethod
    def _at_end_of_headers(header_line: bytes, delimiter: bytes) -> bool:
        """Helper method to check if the parser is at the end of the header section.

        :param header_line: Contents of the header line.
        :param delimiter: Separator between the header line and the rest of the response.

        :return: If the header section is finished and there is a delimiter.
        """

        return all([not header_line, delimiter])

    @staticmethod
    def _at_next_header(header_line: bytes, delimiter: bytes) -> bool:
        """Helper method to check if a header can be parsed.

        :param header_line: Contents of the header line.
        :param delimiter: Separator between the header line and the rest of the response.

        :return: If the header line is finished, and there is another header.
        """

        return all([header_line, delimiter])

    @staticmethod
    def _is_status_line_only(
        header_line: bytes, delimiter: bytes, leftover: bytes
    ) -> bool:
        """Helper method to check if the response is a status line only, without a header section.

        :param header_line: Contents of the header line.
        :param delimiter: Separator between the header line and the rest of the response.
        :param leftover: Remaining response contents.

        :return: If the response was a status line only without a header section or body.
        """

        return all([not header_line, not delimiter, not leftover])

    def body(self) -> None:
        """Uses the length defined in the `Content-length` header (defaulted to 0) to determine how many bytes the body
        contains.

        :raises TooMuchDataError:
            When there are too many bytes in the buffer compared to the `Content-length` header value.  Transitions the
            state to :class:`States.Done`.
        """

        content_length = (
            self.result["headers"]
            .get("Content-length", ContentLengthValue(length=0))
            .length
        )
        try:
            self.result["body"] += self.body_parser(self.buffer, content_length)
            self._state = States.Done
            self._logger.debug("Finished parsing body")
        except TooMuchDataError:
            self._state = States.Done
            self._logger.error("Body parsed was larger than the content length")
            raise


def parse_request_status(stream: bytes) -> dict[str, str]:
    """Parses the status line from a request.

    :param stream: The byte stream to parse.

    :return: A dictionary with the keys `verb`, `protocol` and `version`.

    :raises ParseError:
        When the status line is in an invalid format, not a valid verb, or doesn't have the correct protocol.
    """

    try:
        verb, protocol_version = stream.decode("ascii").split(" ")
        protocol, version = protocol_version.split("/")
    except ValueError as error:
        raise ParseError(
            "Could not parse request status line, not in recognizable format"
        ) from error

    if verb not in [
        "CHECK",
        "HEADERS",
        "PING",
        "PROCESS",
        "REPORT_IFSPAM",
        "REPORT",
        "SKIP",
        "SYMBOLS",
        "TELL",
    ]:
        raise ParseError("Not a valid verb")

    if protocol != "SPAMC":
        raise ParseError("Protocol name does not match")

    return {"verb": verb, "protocol": protocol, "version": version}


def parse_response_status(stream: bytes) -> dict[str, Union[str, int]]:
    """Parse the status line for a response.

    :param stream: The byte stream to parse.

    :return: A dictionary with the keys `protocol`, `version`, `status_code`, and `message`.

    :raises ParseError:
        When the status line is in an invalid format, status code is not an integer, or doesn't have the correct
        protocol.
    """

    try:
        protocol_version, status_code, message = list(
            filter(None, stream.decode("ascii").split(" ", 2))
        )
        protocol, version = protocol_version.split("/")
    except ValueError as error:
        raise ParseError(
            "Could not parse response status line, not in recognizable format"
        ) from error

    if protocol != "SPAMD":
        raise ParseError("Protocol name does not match")

    try:
        status_code_int = int(status_code)
    except ValueError as error:
        raise ParseError("Protocol status code is not an integer") from error

    return {
        "protocol": protocol,
        "version": version,
        "status_code": status_code_int,
        "message": message,
    }


def parse_message_class_value(
    stream: Union[str, MessageClassOption]
) -> MessageClassValue:
    """Parses the `Message-class` header value.

    :param stream: String or :class:`aiospamc.header_values.MessageClassOption` instance.

    :return: A :class:`aiospamc.header_values.MessageClassValue` instance representing the value.

    :raises ParseError: When the value doesn't match either `ham` or `spam`.
    """

    value: MessageClassOption
    if isinstance(stream, MessageClassOption):
        value = stream
    else:
        try:
            value = getattr(MessageClassOption, stream.strip())
        except AttributeError as error:
            raise ParseError("Unable to parse Message-class header value") from error

    return MessageClassValue(value=value)


def parse_content_length_value(stream: Union[str, int]) -> ContentLengthValue:
    """Parses the `Content-length` header value.

    :param stream: String or integer value of the header.

    :return: A :class:`aiospamc.header_values.ContentLengthValue` instance.

    :raises ParseError: When the value cannot be cast to an integer.
    """

    try:
        value = int(stream)
    except ValueError as error:
        raise ParseError(
            "Unable to parse Content-length value, must be integer"
        ) from error

    return ContentLengthValue(length=value)


def parse_compress_value(stream: str) -> CompressValue:
    """Parses a value for the `Compress` header.

    :param stream: String to parse.

    :return: A :class:`aiospamc.header_values.CompressValue` instance.
    """

    return CompressValue(algorithm=stream.strip())


def parse_set_remove_value(stream: Union[ActionOption, str]) -> SetOrRemoveValue:
    """Parse a value for the :class:`aiospamc.header_values.DidRemove`, :class:`aiospamc.header_values.DidSet`, :class:`aiospamc.header_values.Remove`, and :class:`aiospamc.header_values.Set` headers.

    :param stream: String to parse or an instance of :class:`aiospamc.header_values.ActionOption`.

    :return: A :class:`aiospamc.header_values.SetOrRemoveValue` instance.
    """

    if isinstance(stream, ActionOption):
        value = stream
    else:
        stream = stream.replace(" ", "")
        values = stream.split(",")

        if "local" in values:
            local = True
        else:
            local = False

        if "remote" in values:
            remote = True
        else:
            remote = False

        value = ActionOption(local=local, remote=remote)

    return SetOrRemoveValue(action=value)


def parse_spam_value(stream: str) -> SpamValue:
    """Parses the values for the `Spam` header.

    :param stream: String to parse.

    :return: A :class:`aiospamc.header_values.SpamValue` instance.

    :raises ParseError: Raised if there is no true/false value, or valid numbers for the score or threshold.
    """

    stream = stream.replace(" ", "")
    try:
        found, score, threshold = re.split("[;/]", stream)
    except ValueError as error:
        raise ParseError("Spam header in unrecognizable format") from error

    found = found.lower()
    if found in ["true", "yes"]:
        value = True
    elif found in ["false", "no"]:
        value = False
    else:
        raise ParseError("Spam header is not a true or false value")

    try:
        parsed_score = float(score)
    except ValueError as error:
        raise ParseError("Cannot parse Spam header score value") from error

    try:
        parsed_threshold = float(threshold)
    except ValueError as error:
        raise ParseError("Cannot parse Spam header threshold value") from error

    return SpamValue(value=value, score=parsed_score, threshold=parsed_threshold)


def parse_user_value(stream: str) -> UserValue:
    """Parse the username.

    :param stream: String of username to parse. Whitespace is trimmed.

    :return: The :class:`aiospamc.header_values.UserValue` instance.
    """

    return UserValue(name=stream.strip())


def parse_header_value(header: str, value: Union[str, bytes]) -> Any:
    """Sends the header value stream to the header value parsing function.

    :param header: Name of the header.
    :param value: String or byte stream of the header value.

    :return: The :class:`aiospamc.header_values.HeaderValue` instance from the parsing function.
    """

    if header in header_value_parsers:
        if isinstance(value, bytes):
            try:
                return header_value_parsers[header](value.decode())
            except UnicodeDecodeError as error:
                raise ParseError(message="Unable to decode header value") from error
        else:
            return header_value_parsers[header](value)
    else:
        if isinstance(value, bytes):
            try:
                return GenericHeaderValue(value.decode())
            except UnicodeDecodeError:
                return BytesHeaderValue(value)
        else:
            return GenericHeaderValue(value)


def parse_header(stream: bytes) -> tuple[str, Any]:
    """Splits the header line and sends to the header parsing function.

    :param stream: Byte stream of the header line.

    :return: A tuple of the header name and value.
    """

    header, _, value = stream.partition(b":")
    parsed_header = header.decode("ascii").strip()
    parsed_value = parse_header_value(parsed_header, value)

    return parsed_header, parsed_value


def parse_body(stream: bytes, content_length: int) -> bytes:
    """Parses the body of a message.

    :param stream: Byte stream for the body.
    :param content_length: Expected length of the body in bytes.

    :return: Byte stream of the body.

    :raises NotEnoughDataError: If the length is less than the stream.
    :raises TooMuchDataError: If the length is more than the stream.
    """

    if len(stream) == content_length:
        return stream
    elif len(stream) < content_length:
        raise NotEnoughDataError
    else:
        raise TooMuchDataError


header_value_parsers = {
    "Compress": parse_compress_value,
    "Content-length": parse_content_length_value,
    "DidRemove": parse_set_remove_value,
    "DidSet": parse_set_remove_value,
    "Message-class": parse_message_class_value,
    "Remove": parse_set_remove_value,
    "Set": parse_set_remove_value,
    "Spam": parse_spam_value,
    "User": parse_user_value,
}
"""Mapping for header names to their parsing functions."""


class RequestParser(Parser):
    """Sub-class of the parser for requests."""

    def __init__(self):
        """RequestParse constructor."""

        super().__init__(
            delimiter=b"\r\n",
            status_parser=parse_request_status,
            header_parser=parse_header,
            body_parser=parse_body,
        )


class ResponseParser(Parser):
    """Sub-class of the parser for responses."""

    def __init__(self):
        """ResponseParser constructor."""

        super().__init__(
            delimiter=b"\r\n",
            status_parser=parse_response_status,
            header_parser=parse_header,
            body_parser=parse_body,
        )



"""Collection of exceptions."""


class ClientException(Exception):
    """Base class for exceptions raised from the client."""

    pass


class BadRequest(ClientException):
    """Request is not in the expected format."""

    pass


class BadResponse(ClientException):
    """Response is not in the expected format."""

    pass


class AIOSpamcConnectionFailed(ClientException):
    """Connection failed."""

    pass


class TimeoutException(Exception):
    """General timeout exception."""

    pass


class ClientTimeoutException(ClientException, TimeoutException):
    """Timeout exception from the client."""

    pass


class ParseError(Exception):
    """Error occurred while parsing."""

    def __init__(self, message=None):
        """Construct parsing exception with optional message.

        :param message: User friendly message.
        """

        self.message = message


class NotEnoughDataError(ParseError):
    """Expected more data than what the protocol content specified."""

    pass


class TooMuchDataError(ParseError):
    """Too much data was received than what the protocol content specified."""

    pass



"""Contains all requests that can be made to the SPAMD service."""

import zlib
from base64 import b64encode
from typing import Any, SupportsBytes, Union

from .header_values import ContentLengthValue, Headers


class Request:
    """SPAMC request object."""

    def __init__(
        self,
        verb: str,
        version: str = "1.5",
        headers: Union[dict[str, Any], Headers, None] = None,
        body: Union[bytes, SupportsBytes] = b"",
        **_,
    ) -> None:
        """Request constructor.

        :param verb: Method name of the request.
        :param version: Version of the protocol.
        :param headers: Collection of headers to be added.
        :param body: Byte string representation of the body.
        """

        self.verb = verb
        self.version = version
        if isinstance(headers, dict):
            self.headers = Headers(headers)
        elif isinstance(headers, Headers):
            self.headers = headers
        else:
            self.headers = Headers()
        self.body = bytes(body)

    def __bytes__(self) -> bytes:
        if "Compress" in self.headers.keys():
            body = zlib.compress(self.body)
        else:
            body = self.body

        if len(body) > 0:
            self.headers["Content-length"] = ContentLengthValue(length=len(body))

        encoded_headers = b"".join(
            [
                b"%b: %b\r\n" % (key.encode("ascii"), bytes(value))
                for key, value in self.headers.items()
            ]
        )

        request = (
            b"%(verb)b " b"SPAMC/%(version)b" b"\r\n" b"%(headers)b\r\n" b"%(body)b"
        )

        return request % {
            b"verb": self.verb.encode("ascii"),
            b"version": self.version.encode("ascii"),
            b"headers": encoded_headers,
            b"body": body,
        }

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return (
            f"<{self.__class__.__module__}.{self.__class__.__qualname__} "
            f"[{self.verb}] "
            f"object at {hex(id(self))}>"
        )

    @property
    def body(self) -> bytes:
        """Body property getter.

        :return: Value of body.
        """

        return self._body

    @body.setter
    def body(self, value: Union[bytes, SupportsBytes]) -> None:
        """Body property setter.

        :param value: Value to set the body.
        """

        self._body = bytes(value)

    def to_json(self):
        """Converts to JSON serializable object."""

        return {
            "verb": self.verb,
            "version": self.version,
            "headers": {key: value.to_json() for key, value in self.headers.items()},
            "body": b64encode(self.body).decode(),
        }



"""aiospamc package.

An asyncio-based library to communicate with SpamAssassin's SPAMD service."""

from loguru import logger

from .connections import Timeout
from .frontend import (
    check,
    headers,
    ping,
    process,
    report,
    report_if_spam,
    symbols,
    tell,
)
from .header_values import ActionOption, MessageClassOption

__author__ = "Michael Caley"
__copyright__ = "Copyright 2016-2023 Michael Caley"
__license__ = "MIT"
__version__ = "1.1.1"
__email__ = "mjcaley@darkarctic.com"

logger.disable(__package__)
